@@ -1,6 +1,6 @@
 /****************************************************************************
 **
-** Copyright (C) 2016 The Qt Company Ltd.
+** Copyright (C) 2017 The Qt Company Ltd.
 ** Contact: http://www.qt.io/licensing/
 **
 ** This file is part of the Qt Quick Templates 2 module of the Qt Toolkit.
@@ -49,12 +49,14 @@ QT_BEGIN_NAMESPACE
     \inqmlmodule QtQuick.Controls
     \since 5.7
     \ingroup qtquickcontrols2-input
-    \brief A circular dial that is rotated to set a value.
+    \brief Circular dial that is rotated to set a value.
 
     The Dial is similar to a traditional dial knob that is found on devices
     such as stereos or industrial equipment. It allows the user to specify a
     value within a range.
 
+    \image qtquickcontrols2-dial-no-wrap.gif
+
     The value of the dial is set with the \l value property. The range is
     set with the \l from and \l to properties. To enable or disable wrapping,
     use the \l wrap property.
@@ -75,6 +77,14 @@ QT_BEGIN_NAMESPACE
     \sa {Customizing Dial}, {Input Controls}
 */
 
+/*!
+    \since QtQuick.Controls 2.2 (Qt 5.9)
+    \qmlsignal QtQuick.Controls::Dial::moved()
+
+    This signal is emitted when the dial has been interactively moved
+    by the user by either touch, mouse, or keys.
+*/
+
 static const qreal startAngleRadians = (M_PI * 2.0) * (4.0 / 6.0);
 static const qreal startAngle = -140;
 static const qreal endAngleRadians = (M_PI * 2.0) * (5.0 / 6.0);
@@ -85,26 +95,32 @@ class QQuickDialPrivate : public QQuickControlPrivate
     Q_DECLARE_PUBLIC(QQuickDial)
 
 public:
-    QQuickDialPrivate() :
-        from(0),
-        to(1),
-        value(0),
-        position(0),
-        angle(startAngle),
-        stepSize(0),
-        pressed(false),
-        snapMode(QQuickDial::NoSnap),
-        wrap(false),
-        handle(nullptr)
+    QQuickDialPrivate()
+        : from(0),
+          to(1),
+          value(0),
+          position(0),
+          angle(startAngle),
+          stepSize(0),
+          pressed(false),
+          snapMode(QQuickDial::NoSnap),
+          wrap(false),
+          live(true),
+          handle(nullptr)
     {
     }
 
     qreal valueAt(qreal position) const;
     qreal snapPosition(qreal position) const;
-    qreal positionAt(const QPoint &point) const;
+    qreal positionAt(const QPointF &point) const;
     void setPosition(qreal position);
     void updatePosition();
-    bool isLargeChange(const QPoint &eventPos, qreal proposedPosition) const;
+    bool isLargeChange(const QPointF &eventPos, qreal proposedPosition) const;
+
+    void handlePress(const QPointF &point) override;
+    void handleMove(const QPointF &point) override;
+    void handleRelease(const QPointF &point) override;
+    void handleUngrab() override;
 
     qreal from;
     qreal to;
@@ -113,9 +129,10 @@ public:
     qreal angle;
     qreal stepSize;
     bool pressed;
-    QPoint pressPoint;
+    QPointF pressPoint;
     QQuickDial::SnapMode snapMode;
     bool wrap;
+    bool live;
     QQuickItem *handle;
 };
 
@@ -137,7 +154,7 @@ qreal QQuickDialPrivate::snapPosition(qreal position) const
     return qRound(position / effectiveStep) * effectiveStep;
 }
 
-qreal QQuickDialPrivate::positionAt(const QPoint &point) const
+qreal QQuickDialPrivate::positionAt(const QPointF &point) const
 {
     qreal yy = height / 2.0 - point.y();
     qreal xx = point.x() - width / 2.0;
@@ -173,16 +190,77 @@ void QQuickDialPrivate::updatePosition()
     setPosition(pos);
 }
 
-bool QQuickDialPrivate::isLargeChange(const QPoint &eventPos, qreal proposedPosition) const
+bool QQuickDialPrivate::isLargeChange(const QPointF &eventPos, qreal proposedPosition) const
 {
     return qAbs(proposedPosition - position) >= 0.5 && eventPos.y() >= height / 2;
 }
 
-QQuickDial::QQuickDial(QQuickItem *parent) :
-    QQuickControl(*(new QQuickDialPrivate), parent)
+void QQuickDialPrivate::handlePress(const QPointF &point)
+{
+    Q_Q(QQuickDial);
+    QQuickControlPrivate::handlePress(point);
+    pressPoint = point;
+    q->setPressed(true);
+}
+
+void QQuickDialPrivate::handleMove(const QPointF &point)
+{
+    Q_Q(QQuickDial);
+    QQuickControlPrivate::handleMove(point);
+    const qreal oldPos = position;
+    qreal pos = positionAt(point);
+    if (snapMode == QQuickDial::SnapAlways)
+        pos = snapPosition(pos);
+
+    if (wrap || (!wrap && !isLargeChange(point, pos))) {
+        if (live)
+            q->setValue(valueAt(pos));
+        else
+            setPosition(pos);
+        if (!qFuzzyCompare(pos, oldPos))
+            emit q->moved();
+    }
+}
+
+void QQuickDialPrivate::handleRelease(const QPointF &point)
+{
+    Q_Q(QQuickDial);
+    QQuickControlPrivate::handleRelease(point);
+    if (q->keepMouseGrab() || q->keepTouchGrab()) {
+        const qreal oldPos = position;
+        qreal pos = positionAt(point);
+        if (snapMode != QQuickDial::NoSnap)
+            pos = snapPosition(pos);
+
+        if (wrap || (!wrap && !isLargeChange(point, pos)))
+            q->setValue(valueAt(pos));
+        if (!qFuzzyCompare(pos, oldPos))
+            emit q->moved();
+
+        q->setKeepMouseGrab(false);
+        q->setKeepTouchGrab(false);
+    }
+
+    q->setPressed(false);
+    pressPoint = QPointF();
+}
+
+void QQuickDialPrivate::handleUngrab()
+{
+    Q_Q(QQuickDial);
+    QQuickControlPrivate::handleUngrab();
+    pressPoint = QPointF();
+    q->setPressed(false);
+}
+
+QQuickDial::QQuickDial(QQuickItem *parent)
+    : QQuickControl(*(new QQuickDialPrivate), parent)
 {
     setActiveFocusOnTab(true);
     setAcceptedMouseButtons(Qt::LeftButton);
+#if QT_CONFIG(cursor)
+    setCursor(Qt::ArrowCursor);
+#endif
 }
 
 /*!
@@ -246,11 +324,7 @@ void QQuickDial::setTo(qreal to)
     This property holds the value in the range \c from - \c to. The default
     value is \c 0.0.
 
-    Unlike the \l position property, the \c value is not updated while the
-    handle is dragged. The value is updated after the value has been chosen
-    and the dial has been released.
-
-    \sa position
+    \sa position, live
 */
 qreal QQuickDial::value() const
 {
@@ -278,10 +352,8 @@ void QQuickDial::setValue(qreal value)
 
     This property holds the logical position of the handle.
 
-    The position is defined as a percentage of the control's angle range (the
-    range within which the handle can be moved) scaled to \c {0.0 - 1.0}.
-    Unlike the \l value property, the \c position is continuously updated while
-    the handle is dragged.
+    The position is expressed as a fraction of the control's angle range (the
+    range within which the handle can be moved) in the range \c {0.0 - 1.0}.
 
     \sa value, angle
 */
@@ -297,9 +369,6 @@ qreal QQuickDial::position() const
 
     This property holds the angle of the handle.
 
-    Like the \l position property, angle is continuously updated while the
-    handle is dragged.
-
     The range is from \c -140 degrees to \c 140 degrees.
 
     \sa position
@@ -444,6 +513,33 @@ void QQuickDial::setPressed(bool pressed)
     emit pressedChanged();
 }
 
+/*!
+    \since QtQuick.Controls 2.2 (Qt 5.9)
+    \qmlproperty bool QtQuick.Controls::Dial::live
+
+    This property holds whether the dial provides live updates for the \l value
+    property while the handle is dragged.
+
+    The default value is \c true.
+
+    \sa value
+*/
+bool QQuickDial::live() const
+{
+    Q_D(const QQuickDial);
+    return d->live;
+}
+
+void QQuickDial::setLive(bool live)
+{
+    Q_D(QQuickDial);
+    if (d->live == live)
+        return;
+
+    d->live = live;
+    emit liveChanged();
+}
+
 /*!
     \qmlmethod void QtQuick.Controls::Dial::increase()
 
@@ -493,7 +589,7 @@ void QQuickDial::setHandle(QQuickItem *handle)
     if (handle == d->handle)
         return;
 
-    d->deleteDelegate(d->handle);
+    QQuickControlPrivate::destroyDelegate(d->handle, this);
     d->handle = handle;
     if (d->handle && !d->handle->parentItem())
         d->handle->setParentItem(this);
@@ -503,6 +599,7 @@ void QQuickDial::setHandle(QQuickItem *handle)
 void QQuickDial::keyPressEvent(QKeyEvent *event)
 {
     Q_D(QQuickDial);
+    const qreal oldValue = d->value;
     switch (event->key()) {
     case Qt::Key_Left:
     case Qt::Key_Down:
@@ -537,6 +634,8 @@ void QQuickDial::keyPressEvent(QKeyEvent *event)
         QQuickControl::keyPressEvent(event);
         break;
     }
+    if (!qFuzzyCompare(d->value, oldValue))
+        emit moved();
 }
 
 void QQuickDial::keyReleaseEvent(QKeyEvent *event)
@@ -549,60 +648,78 @@ void QQuickDial::mousePressEvent(QMouseEvent *event)
 {
     Q_D(QQuickDial);
     QQuickControl::mousePressEvent(event);
-    d->pressPoint = event->pos();
-    setPressed(true);
+    d->handleMove(event->localPos());
 }
 
 void QQuickDial::mouseMoveEvent(QMouseEvent *event)
 {
     Q_D(QQuickDial);
-    QQuickControl::mouseMoveEvent(event);
     if (!keepMouseGrab()) {
-        bool overXDragThreshold = QQuickWindowPrivate::dragOverThreshold(event->pos().x() - d->pressPoint.x(), Qt::XAxis, event);
+        bool overXDragThreshold = QQuickWindowPrivate::dragOverThreshold(event->localPos().x() - d->pressPoint.x(), Qt::XAxis, event);
         setKeepMouseGrab(overXDragThreshold);
 
         if (!overXDragThreshold) {
-            bool overYDragThreshold = QQuickWindowPrivate::dragOverThreshold(event->pos().y() - d->pressPoint.y(), Qt::YAxis, event);
+            bool overYDragThreshold = QQuickWindowPrivate::dragOverThreshold(event->localPos().y() - d->pressPoint.y(), Qt::YAxis, event);
             setKeepMouseGrab(overYDragThreshold);
         }
     }
-    if (keepMouseGrab()) {
-        qreal pos = d->positionAt(event->pos());
-        if (d->snapMode == SnapAlways)
-            pos = d->snapPosition(pos);
-
-        if (d->wrap || (!d->wrap && !d->isLargeChange(event->pos(), pos)))
-            d->setPosition(pos);
-    }
+    QQuickControl::mouseMoveEvent(event);
 }
 
-void QQuickDial::mouseReleaseEvent(QMouseEvent *event)
+#if QT_CONFIG(quicktemplates2_multitouch)
+void QQuickDial::touchEvent(QTouchEvent *event)
 {
     Q_D(QQuickDial);
-    QQuickControl::mouseReleaseEvent(event);
+    switch (event->type()) {
+    case QEvent::TouchUpdate:
+        for (const QTouchEvent::TouchPoint &point : event->touchPoints()) {
+            if (!d->acceptTouch(point))
+                continue;
 
-    if (keepMouseGrab()) {
-        qreal pos = d->positionAt(event->pos());
-        if (d->snapMode != NoSnap)
-            pos = d->snapPosition(pos);
+            switch (point.state()) {
+            case Qt::TouchPointMoved:
+                if (!keepTouchGrab()) {
+                    bool overXDragThreshold = QQuickWindowPrivate::dragOverThreshold(point.pos().x() - d->pressPoint.x(), Qt::XAxis, &point);
+                    setKeepTouchGrab(overXDragThreshold);
 
-        if (d->wrap || (!d->wrap && !d->isLargeChange(event->pos(), pos)))
-            setValue(d->valueAt(pos));
+                    if (!overXDragThreshold) {
+                        bool overYDragThreshold = QQuickWindowPrivate::dragOverThreshold(point.pos().y() - d->pressPoint.y(), Qt::YAxis, &point);
+                        setKeepTouchGrab(overYDragThreshold);
+                    }
+                }
+                if (keepTouchGrab())
+                    d->handleMove(point.pos());
+                break;
 
-        setKeepMouseGrab(false);
+            default:
+                QQuickControl::touchEvent(event);
+                break;
+            }
+        }
+        break;
+
+    default:
+        QQuickControl::touchEvent(event);
+        break;
     }
-
-    setPressed(false);
-    d->pressPoint = QPoint();
 }
+#endif
 
-void QQuickDial::mouseUngrabEvent()
+#if QT_CONFIG(wheelevent)
+void QQuickDial::wheelEvent(QWheelEvent *event)
 {
     Q_D(QQuickDial);
-    QQuickControl::mouseUngrabEvent();
-    d->pressPoint = QPoint();
-    setPressed(false);
+    QQuickControl::wheelEvent(event);
+    if (d->wheelEnabled) {
+        const qreal oldValue = d->value;
+        const QPointF angle = event->angleDelta();
+        const qreal delta = (qFuzzyIsNull(angle.y()) ? angle.x() : (event->inverted() ? -angle.y() : angle.y())) / QWheelEvent::DefaultDeltasPerStep;
+        const qreal step = qFuzzyIsNull(d->stepSize) ? 0.1 : d->stepSize;
+        setValue(oldValue + step * delta);
+        event->setAccepted(!qFuzzyCompare(d->value, oldValue));
+    }
 }
+#endif
 
 void QQuickDial::mirrorChange()
 {
@@ -618,7 +735,7 @@ void QQuickDial::componentComplete()
     d->updatePosition();
 }
 
-#ifndef QT_NO_ACCESSIBILITY
+#if QT_CONFIG(accessibility)
 void QQuickDial::accessibilityActiveChanged(bool active)
 {
     QQuickControl::accessibilityActiveChanged(active);
