@@ -1,6 +1,6 @@
 /****************************************************************************
 **
-** Copyright (C) 2016 The Qt Company Ltd.
+** Copyright (C) 2017 The Qt Company Ltd.
 ** Contact: http://www.qt.io/licensing/
 **
 ** This file is part of the Qt Quick Templates 2 module of the Qt Toolkit.
@@ -36,6 +36,7 @@
 
 #include "qquickdrawer_p.h"
 #include "qquickdrawer_p_p.h"
+#include "qquickpopupitem_p_p.h"
 
 #include <QtGui/qstylehints.h>
 #include <QtGui/private/qguiapplication_p.h>
@@ -53,21 +54,17 @@ QT_BEGIN_NAMESPACE
     \since 5.7
     \ingroup qtquickcontrols2-navigation
     \ingroup qtquickcontrols2-popups
-    \brief Provides a swipe-based side panel.
+    \brief Side panel that can be opened and closed using a swipe gesture.
 
     Drawer provides a swipe-based side panel, similar to those often used in
     touch interfaces to provide a central location for navigation.
 
-    \table
-    \row
-        \li \image qtquickcontrols2-drawer-wireframe.png
-            Drawer can be positioned at any of the four edges of the content item. \br
-            In this image, it is against the left edge of the window.
+    \image qtquickcontrols2-drawer.gif
 
-        \li \image qtquickcontrols2-drawer-expanded-wireframe.png
-            The drawer is then opened by \e "dragging" it out from the left edge \br
-            of the window.
-    \endtable
+    Drawer can be positioned at any of the four edges of the content item.
+    The drawer above is positioned against the left edge of the window. The
+    drawer is then opened by \e "dragging" it out from the left edge of the
+    window.
 
     \code
     import QtQuick 2.7
@@ -122,7 +119,7 @@ QT_BEGIN_NAMESPACE
 
     \code
     import QtQuick 2.7
-    import QtQuick.Controls 2.0
+    import QtQuick.Controls 2.1
 
     ApplicationWindow {
         id: window
@@ -156,18 +153,33 @@ QT_BEGIN_NAMESPACE
     the drawer is opened, don't apply a translation.
 
     \note On some platforms, certain edges may be reserved for system
-    gestures and therefore cannot be used with Drawer.
+    gestures and therefore cannot be used with Drawer. For example, the
+    top and bottom edges may be reserved for system notifications and
+    control centers on Android and iOS.
 
     \sa SwipeView, {Customizing Drawer}, {Navigation Controls}, {Popup Controls}
 */
 
 QQuickDrawerPrivate::QQuickDrawerPrivate()
-    : edge(Qt::LeftEdge), offset(0), position(0),
+    : edge(Qt::LeftEdge),
+      offset(0),
+      position(0),
       dragMargin(QGuiApplication::styleHints()->startDragDistance())
 {
     setEdge(Qt::LeftEdge);
 }
 
+qreal QQuickDrawerPrivate::offsetAt(const QPointF &point) const
+{
+    qreal offset = positionAt(point) - position;
+
+    // don't jump when dragged open
+    if (offset > 0 && position > 0 && !contains(point))
+        offset = 0;
+
+    return offset;
+}
+
 qreal QQuickDrawerPrivate::positionAt(const QPointF &point) const
 {
     Q_Q(const QQuickDrawer);
@@ -233,47 +245,65 @@ void QQuickDrawerPrivate::resizeOverlay()
     dimmer->setSize(geometry.size());
 }
 
-static bool dragOverThreshold(qreal d, Qt::Axis axis, QMouseEvent *event, int threshold = -1)
+static bool isWithinDragMargin(QQuickDrawer *drawer, const QPointF &pos)
 {
-    return QQuickWindowPrivate::dragOverThreshold(d, axis, event, threshold);
+    switch (drawer->edge()) {
+    case Qt::LeftEdge:
+        return pos.x() <= drawer->dragMargin();
+    case Qt::RightEdge:
+        return pos.x() >= drawer->window()->width() - drawer->dragMargin();
+    case Qt::TopEdge:
+        return pos.y() <= drawer->dragMargin();
+    case Qt::BottomEdge:
+        return pos.y() >= drawer->window()->height() - drawer->dragMargin();
+    default:
+        Q_UNREACHABLE();
+        break;
+    }
+    return false;
 }
 
-bool QQuickDrawerPrivate::startDrag(QQuickWindow *window, QMouseEvent *event)
+bool QQuickDrawerPrivate::startDrag(QEvent *event)
 {
-    if (!window || dragMargin < 0.0 || qFuzzyIsNull(dragMargin))
+    Q_Q(QQuickDrawer);
+    if (!window || !interactive || dragMargin < 0.0 || qFuzzyIsNull(dragMargin))
         return false;
 
-    bool drag = false;
-    switch (edge) {
-    case Qt::LeftEdge:
-        drag = !dragOverThreshold(event->windowPos().x(), Qt::XAxis, event, dragMargin);
+    switch (event->type()) {
+    case QEvent::MouseButtonPress:
+        if (isWithinDragMargin(q, static_cast<QMouseEvent *>(event)->windowPos())) {
+            prepareEnterTransition();
+            reposition();
+            return handleMouseEvent(window->contentItem(), static_cast<QMouseEvent *>(event));
+        }
         break;
-    case Qt::RightEdge:
-        drag = !dragOverThreshold(window->width() - event->windowPos().x(), Qt::XAxis, event, dragMargin);
-        break;
-    case Qt::TopEdge:
-        drag = !dragOverThreshold(event->windowPos().y(), Qt::YAxis, event, dragMargin);
-        break;
-    case Qt::BottomEdge:
-        drag = !dragOverThreshold(window->height() - event->windowPos().y(), Qt::YAxis, event, dragMargin);
+
+#if QT_CONFIG(quicktemplates2_multitouch)
+    case QEvent::TouchBegin:
+    case QEvent::TouchUpdate:
+        for (const QTouchEvent::TouchPoint &point : static_cast<QTouchEvent *>(event)->touchPoints()) {
+            if (point.state() == Qt::TouchPointPressed && isWithinDragMargin(q, point.scenePos())) {
+                prepareEnterTransition();
+                reposition();
+                return handleTouchEvent(window->contentItem(), static_cast<QTouchEvent *>(event));
+            }
+        }
         break;
+#endif
+
     default:
         break;
     }
 
-    if (drag) {
-        prepareEnterTransition();
-        reposition();
-        handleMousePressEvent(window->contentItem(), event);
-    }
-
-    return drag;
+    return false;
 }
 
-bool QQuickDrawerPrivate::grabMouse(QMouseEvent *event)
+bool QQuickDrawerPrivate::grabMouse(QQuickItem *item, QMouseEvent *event)
 {
     Q_Q(QQuickDrawer);
-    if (!window || popupItem->keepMouseGrab())
+    handleMouseEvent(item, event);
+
+    if (!window || !interactive || popupItem->keepMouseGrab() || popupItem->keepTouchGrab())
         return false;
 
     const QPointF movePoint = event->windowPos();
@@ -284,141 +314,188 @@ bool QQuickDrawerPrivate::grabMouse(QMouseEvent *event)
     const int threshold = qMax(20, QGuiApplication::styleHints()->startDragDistance() + 5);
     bool overThreshold = false;
     if (position > 0 || dragMargin > 0) {
+        const bool xOverThreshold = QQuickWindowPrivate::dragOverThreshold(movePoint.x() - pressPoint.x(), Qt::XAxis, event, threshold);
+        const bool yOverThreshold = QQuickWindowPrivate::dragOverThreshold(movePoint.y() - pressPoint.y(), Qt::YAxis, event, threshold);
         if (edge == Qt::LeftEdge || edge == Qt::RightEdge)
-            overThreshold = dragOverThreshold(movePoint.x() - pressPoint.x(), Qt::XAxis, event, threshold);
+            overThreshold = xOverThreshold && !yOverThreshold;
         else
-            overThreshold = dragOverThreshold(movePoint.y() - pressPoint.y(), Qt::YAxis, event, threshold);
+            overThreshold = yOverThreshold && !xOverThreshold;
     }
 
     // Don't be too eager to steal presses outside the drawer (QTBUG-53929)
-    if (overThreshold && qFuzzyCompare(position, qreal(1.0)) && !popupItem->contains(popupItem->mapFromScene(movePoint))) {
+    if (overThreshold && qFuzzyCompare(position, qreal(1.0)) && !contains(movePoint)) {
         if (edge == Qt::LeftEdge || edge == Qt::RightEdge)
             overThreshold = qAbs(movePoint.x() - q->width()) < dragMargin;
         else
             overThreshold = qAbs(movePoint.y() - q->height()) < dragMargin;
     }
 
-    return overThreshold;
-}
-
-static const qreal openCloseVelocityThreshold = 300;
-
-bool QQuickDrawerPrivate::ungrabMouse(QMouseEvent *event)
-{
-    bool wasGrabbed = popupItem->keepMouseGrab();
-    if (wasGrabbed) {
-        const QPointF releasePoint = event->windowPos();
-        velocityCalculator.stopMeasuring(releasePoint, event->timestamp());
-
-        qreal velocity = 0;
-        if (edge == Qt::LeftEdge || edge == Qt::RightEdge)
-            velocity = velocityCalculator.velocity().x();
-        else
-            velocity = velocityCalculator.velocity().y();
-
-        // the velocity is calculated so that swipes from left to right
-        // and top to bottom have positive velocity, and swipes from right
-        // to left and bottom to top have negative velocity.
-        //
-        // - top/left edge: positive velocity opens, negative velocity closes
-        // - bottom/right edge: negative velocity opens, positive velocity closes
-        //
-        // => invert the velocity for bottom and right edges, for the threshold comparison below
-        if (edge == Qt::RightEdge || edge == Qt::BottomEdge)
-            velocity = -velocity;
-
-        if (position > 0.7 || velocity > openCloseVelocityThreshold) {
-            transitionManager.transitionEnter();
-        } else if (position < 0.3 || velocity < -openCloseVelocityThreshold) {
-            transitionManager.transitionExit();
-        } else {
-            switch (edge) {
-            case Qt::LeftEdge:
-                if (releasePoint.x() - pressPoint.x() > 0)
-                    transitionManager.transitionEnter();
-                else
-                    transitionManager.transitionExit();
-                break;
-            case Qt::RightEdge:
-                if (releasePoint.x() - pressPoint.x() < 0)
-                    transitionManager.transitionEnter();
-                else
-                    transitionManager.transitionExit();
-                break;
-            case Qt::TopEdge:
-                if (releasePoint.y() - pressPoint.y() > 0)
-                    transitionManager.transitionEnter();
-                else
-                    transitionManager.transitionExit();
-                break;
-            case Qt::BottomEdge:
-                if (releasePoint.y() - pressPoint.y() < 0)
-                    transitionManager.transitionEnter();
-                else
-                    transitionManager.transitionExit();
-                break;
-            }
-        }
-    }
-    return wasGrabbed;
-}
-
-bool QQuickDrawerPrivate::handleMousePressEvent(QQuickItem *item, QMouseEvent *event)
-{
-    offset = 0;
-    pressPoint = event->windowPos();
-    velocityCalculator.startMeasuring(pressPoint, event->timestamp());
-
-    // don't block press events
-    // a) outside a non-modal drawer,
-    // b) to drawer children, or
-    // c) outside a modal drawer's background dimming
-    event->setAccepted(modal && !popupItem->isAncestorOf(item) && (!dimmer || dimmer->contains(dimmer->mapFromScene(pressPoint))));
-    return event->isAccepted();
-}
-
-bool QQuickDrawerPrivate::handleMouseMoveEvent(QQuickItem *item, QMouseEvent *event)
-{
-    Q_Q(QQuickDrawer);
-    Q_UNUSED(item);
-
-    // Don't react to synthesized mouse move events at INF,INF coordinates.
-    // QQuickWindowPrivate::translateTouchToMouse() uses them to clear hover
-    // on touch release (QTBUG-55995).
-    if (qIsInf(event->screenPos().x()) || qIsInf(event->screenPos().y()))
-        return true;
-
-    const QPointF movePoint = event->windowPos();
-
-    if (grabMouse(event)) {
+    if (overThreshold) {
         QQuickItem *grabber = window->mouseGrabberItem();
         if (!grabber || !grabber->keepMouseGrab()) {
             popupItem->grabMouse();
             popupItem->setKeepMouseGrab(true);
-            offset = qMin<qreal>(0.0, positionAt(movePoint) - position);
+            offset = offsetAt(movePoint);
         }
     }
 
-    if (popupItem->keepMouseGrab())
-        q->setPosition(positionAt(movePoint) - offset);
-    event->accept();
-
-    return popupItem->keepMouseGrab();
+    return overThreshold;
 }
 
-bool QQuickDrawerPrivate::handleMouseReleaseEvent(QQuickItem *item, QMouseEvent *event)
+#if QT_CONFIG(quicktemplates2_multitouch)
+bool QQuickDrawerPrivate::grabTouch(QQuickItem *item, QTouchEvent *event)
 {
-    Q_UNUSED(item);
+    Q_Q(QQuickDrawer);
+    handleTouchEvent(item, event);
 
-    const bool wasGrabbed = ungrabMouse(event);
+    if (!window || !interactive || popupItem->keepTouchGrab() || !event->touchPointStates().testFlag(Qt::TouchPointMoved))
+        return false;
 
+    bool overThreshold = false;
+    for (const QTouchEvent::TouchPoint &point : event->touchPoints()) {
+        if (!acceptTouch(point) || point.state() != Qt::TouchPointMoved)
+            continue;
+
+        const QPointF movePoint = point.scenePos();
+
+        // Flickable uses a hard-coded threshold of 15 for flicking, and
+        // QStyleHints::startDragDistance for dragging. Drawer uses a bit
+        // larger threshold to avoid being too eager to steal touch (QTBUG-50045)
+        const int threshold = qMax(20, QGuiApplication::styleHints()->startDragDistance() + 5);
+        if (position > 0 || dragMargin > 0) {
+            const bool xOverThreshold = QQuickWindowPrivate::dragOverThreshold(movePoint.x() - pressPoint.x(), Qt::XAxis, &point, threshold);
+            const bool yOverThreshold = QQuickWindowPrivate::dragOverThreshold(movePoint.y() - pressPoint.y(), Qt::YAxis, &point, threshold);
+            if (edge == Qt::LeftEdge || edge == Qt::RightEdge)
+                overThreshold = xOverThreshold && !yOverThreshold;
+            else
+                overThreshold = yOverThreshold && !xOverThreshold;
+        }
+
+        // Don't be too eager to steal presses outside the drawer (QTBUG-53929)
+        if (overThreshold && qFuzzyCompare(position, qreal(1.0)) && !contains(movePoint)) {
+            if (edge == Qt::LeftEdge || edge == Qt::RightEdge)
+                overThreshold = qAbs(movePoint.x() - q->width()) < dragMargin;
+            else
+                overThreshold = qAbs(movePoint.y() - q->height()) < dragMargin;
+        }
+
+        if (overThreshold) {
+            popupItem->grabTouchPoints(QVector<int>() << touchId);
+            popupItem->setKeepTouchGrab(true);
+            offset = offsetAt(movePoint);
+        }
+    }
+
+    return overThreshold;
+}
+#endif
+
+static const qreal openCloseVelocityThreshold = 300;
+
+bool QQuickDrawerPrivate::handlePress(QQuickItem *item, const QPointF &point, ulong timestamp)
+{
+    offset = 0;
+    velocityCalculator.startMeasuring(point, timestamp);
+
+    if (!QQuickPopupPrivate::handlePress(item, point, timestamp))
+        return false;
+
+    return true;
+}
+
+bool QQuickDrawerPrivate::handleMove(QQuickItem *item, const QPointF &point, ulong timestamp)
+{
+    Q_Q(QQuickDrawer);
+    if (!QQuickPopupPrivate::handleMove(item, point, timestamp))
+        return false;
+
+    // limit/reset the offset to the edge of the drawer when pushed from the outside
+    if (qFuzzyCompare(position, 1.0) && !contains(point))
+        offset = 0;
+
+    bool isGrabbed = popupItem->keepMouseGrab() || popupItem->keepTouchGrab();
+    if (isGrabbed)
+        q->setPosition(positionAt(point) - offset);
+
+    return isGrabbed;
+}
+
+bool QQuickDrawerPrivate::handleRelease(QQuickItem *item, const QPointF &point, ulong timestamp)
+{
+    if (!popupItem->keepMouseGrab() && !popupItem->keepTouchGrab()) {
+        velocityCalculator.reset();
+        return QQuickPopupPrivate::handleRelease(item, point, timestamp);
+    }
+
+    velocityCalculator.stopMeasuring(point, timestamp);
+
+    qreal velocity = 0;
+    if (edge == Qt::LeftEdge || edge == Qt::RightEdge)
+        velocity = velocityCalculator.velocity().x();
+    else
+        velocity = velocityCalculator.velocity().y();
+
+    // the velocity is calculated so that swipes from left to right
+    // and top to bottom have positive velocity, and swipes from right
+    // to left and bottom to top have negative velocity.
+    //
+    // - top/left edge: positive velocity opens, negative velocity closes
+    // - bottom/right edge: negative velocity opens, positive velocity closes
+    //
+    // => invert the velocity for bottom and right edges, for the threshold comparison below
+    if (edge == Qt::RightEdge || edge == Qt::BottomEdge)
+        velocity = -velocity;
+
+    if (position > 0.7 || velocity > openCloseVelocityThreshold) {
+        transitionManager.transitionEnter();
+    } else if (position < 0.3 || velocity < -openCloseVelocityThreshold) {
+        transitionManager.transitionExit();
+    } else {
+        switch (edge) {
+        case Qt::LeftEdge:
+            if (point.x() - pressPoint.x() > 0)
+                transitionManager.transitionEnter();
+            else
+                transitionManager.transitionExit();
+            break;
+        case Qt::RightEdge:
+            if (point.x() - pressPoint.x() < 0)
+                transitionManager.transitionEnter();
+            else
+                transitionManager.transitionExit();
+            break;
+        case Qt::TopEdge:
+            if (point.y() - pressPoint.y() > 0)
+                transitionManager.transitionEnter();
+            else
+                transitionManager.transitionExit();
+            break;
+        case Qt::BottomEdge:
+            if (point.y() - pressPoint.y() < 0)
+                transitionManager.transitionEnter();
+            else
+                transitionManager.transitionExit();
+            break;
+        }
+    }
+
+    bool wasGrabbed = popupItem->keepMouseGrab() || popupItem->keepTouchGrab();
     popupItem->setKeepMouseGrab(false);
-    pressPoint = QPoint();
-    event->accept();
+    popupItem->setKeepTouchGrab(false);
+
+    pressPoint = QPointF();
+    touchId = -1;
 
     return wasGrabbed;
 }
 
+void QQuickDrawerPrivate::handleUngrab()
+{
+    QQuickPopupPrivate::handleUngrab();
+
+    velocityCalculator.reset();
+}
+
 static QList<QQuickStateAction> prepareTransition(QQuickDrawer *drawer, QQuickTransition *transition, qreal to)
 {
     QList<QQuickStateAction> actions;
@@ -469,8 +546,8 @@ void QQuickDrawerPrivate::setEdge(Qt::Edge e)
     }
 }
 
-QQuickDrawer::QQuickDrawer(QObject *parent) :
-    QQuickPopup(*(new QQuickDrawerPrivate), parent)
+QQuickDrawer::QQuickDrawer(QObject *parent)
+    : QQuickPopup(*(new QQuickDrawerPrivate), parent)
 {
     setFocus(true);
     setModal(true);
@@ -481,13 +558,13 @@ QQuickDrawer::QQuickDrawer(QObject *parent) :
 /*!
     \qmlproperty enumeration QtQuick.Controls::Drawer::edge
 
-    This property holds the edge of the content item at which the drawer will
+    This property holds the edge of the window at which the drawer will
     open from. The acceptable values are:
 
-    \value Qt.TopEdge     The top edge of the content item.
-    \value Qt.LeftEdge    The left edge of the content item (default).
-    \value Qt.RightEdge   The right edge of the content item.
-    \value Qt.BottomEdge  The bottom edge of the content item.
+    \value Qt.TopEdge     The top edge of the window.
+    \value Qt.LeftEdge    The left edge of the window (default).
+    \value Qt.RightEdge   The right edge of the window.
+    \value Qt.BottomEdge  The bottom edge of the window.
 */
 Qt::Edge QQuickDrawer::edge() const
 {
@@ -511,8 +588,8 @@ void QQuickDrawer::setEdge(Qt::Edge edge)
     \qmlproperty real QtQuick.Controls::Drawer::position
 
     This property holds the position of the drawer relative to its final
-    destination. That is, the position will be \c 0 when the drawer
-    is fully closed, and \c 1 when fully open.
+    destination. That is, the position will be \c 0.0 when the drawer
+    is fully closed, and \c 1.0 when fully open.
 */
 qreal QQuickDrawer::position() const
 {
@@ -543,6 +620,8 @@ void QQuickDrawer::setPosition(qreal position)
     prevents opening the drawer by dragging.
 
     The default value is \c Qt.styleHints.startDragDistance.
+
+    \sa interactive
 */
 qreal QQuickDrawer::dragMargin() const
 {
@@ -565,67 +644,83 @@ void QQuickDrawer::resetDragMargin()
     setDragMargin(QGuiApplication::styleHints()->startDragDistance());
 }
 
+/*!
+    \since QtQuick.Controls 2.2 (Qt 5.9)
+    \qmlproperty bool QtQuick.Controls::Drawer::interactive
+
+    This property holds whether the drawer is interactive. A non-interactive
+    drawer does not react to swipes.
+
+    The default value is \c true.
+
+    \sa dragMargin
+*/
+bool QQuickDrawer::isInteractive() const
+{
+    Q_D(const QQuickDrawer);
+    return d->interactive;
+}
+
+void QQuickDrawer::setInteractive(bool interactive)
+{
+    Q_D(QQuickDrawer);
+    if (d->interactive == interactive)
+        return;
+
+    setFiltersChildMouseEvents(interactive);
+    d->interactive = interactive;
+    emit interactiveChanged();
+}
+
 bool QQuickDrawer::childMouseEventFilter(QQuickItem *child, QEvent *event)
 {
     Q_D(QQuickDrawer);
     switch (event->type()) {
-    case QEvent::MouseButtonPress:
-        return d->handleMousePressEvent(child, static_cast<QMouseEvent *>(event));
+#if QT_CONFIG(quicktemplates2_multitouch)
+    case QEvent::TouchUpdate:
+        return d->grabTouch(child, static_cast<QTouchEvent *>(event));
+#endif
     case QEvent::MouseMove:
-        return d->handleMouseMoveEvent(child, static_cast<QMouseEvent *>(event));
+        return d->grabMouse(child, static_cast<QMouseEvent *>(event));
+    case QEvent::MouseButtonPress:
     case QEvent::MouseButtonRelease:
-        return d->handleMouseReleaseEvent(child, static_cast<QMouseEvent *>(event));
+        return d->handleMouseEvent(child, static_cast<QMouseEvent *>(event));
     default:
-        return false;
+        break;
     }
-}
-
-void QQuickDrawer::mousePressEvent(QMouseEvent *event)
-{
-    Q_D(QQuickDrawer);
-    QQuickPopup::mousePressEvent(event);
-    d->handleMousePressEvent(d->popupItem, event);
+    return false;
 }
 
 void QQuickDrawer::mouseMoveEvent(QMouseEvent *event)
 {
     Q_D(QQuickDrawer);
-    QQuickPopup::mouseMoveEvent(event);
-    d->handleMouseMoveEvent(d->popupItem, event);
-}
-
-void QQuickDrawer::mouseReleaseEvent(QMouseEvent *event)
-{
-    Q_D(QQuickDrawer);
-    QQuickPopup::mouseReleaseEvent(event);
-    d->handleMouseReleaseEvent(d->popupItem, event);
-}
-
-void QQuickDrawer::mouseUngrabEvent()
-{
-    Q_D(QQuickDrawer);
-    QQuickPopup::mouseUngrabEvent();
-    d->pressPoint = QPoint();
-    d->velocityCalculator.reset();
+    d->grabMouse(d->popupItem, event);
 }
 
 bool QQuickDrawer::overlayEvent(QQuickItem *item, QEvent *event)
 {
     Q_D(QQuickDrawer);
     switch (event->type()) {
-    case QEvent::MouseButtonPress:
-        d->tryClose(item, static_cast<QMouseEvent *>(event));
-        return d->handleMousePressEvent(item, static_cast<QMouseEvent *>(event));
+#if QT_CONFIG(quicktemplates2_multitouch)
+    case QEvent::TouchUpdate:
+        return d->grabTouch(item, static_cast<QTouchEvent *>(event));
+#endif
     case QEvent::MouseMove:
-        return d->handleMouseMoveEvent(item, static_cast<QMouseEvent *>(event));
-    case QEvent::MouseButtonRelease:
-        d->tryClose(item, static_cast<QMouseEvent *>(event));
-        return d->handleMouseReleaseEvent(item, static_cast<QMouseEvent *>(event));
+        return d->grabMouse(item, static_cast<QMouseEvent *>(event));
     default:
-        return QQuickPopup::overlayEvent(item, event);
+        break;
     }
+    return QQuickPopup::overlayEvent(item, event);
 }
 
+#if QT_CONFIG(quicktemplates2_multitouch)
+void QQuickDrawer::touchEvent(QTouchEvent *event)
+{
+    Q_D(QQuickDrawer);
+    d->grabTouch(d->popupItem, event);
+}
+#endif
+
 void QQuickDrawer::geometryChanged(const QRectF &newGeometry, const QRectF &oldGeometry)
 {
     Q_D(QQuickDrawer);
