@@ -1,6 +1,6 @@
 /****************************************************************************
 **
-** Copyright (C) 2016 The Qt Company Ltd.
+** Copyright (C) 2017 The Qt Company Ltd.
 ** Contact: http://www.qt.io/licensing/
 **
 ** This file is part of the Qt Quick Templates 2 module of the Qt Toolkit.
@@ -37,6 +37,7 @@
 #include "qquickmenu_p.h"
 #include "qquickmenu_p_p.h"
 #include "qquickmenuitem_p.h"
+#include "qquickcontrol_p_p.h"
 
 #include <QtGui/qevent.h>
 #include <QtQml/private/qqmlobjectmodel_p.h>
@@ -101,9 +102,9 @@ QT_BEGIN_NAMESPACE
     \sa {Customizing Menu}, {Menu Controls}, {Popup Controls}
 */
 
-QQuickMenuPrivate::QQuickMenuPrivate() :
-    contentItem(nullptr),
-    contentModel(nullptr)
+QQuickMenuPrivate::QQuickMenuPrivate()
+    : contentItem(nullptr),
+      contentModel(nullptr)
 {
     Q_Q(QQuickMenu);
     contentModel = new QQmlObjectModel(q);
@@ -196,10 +197,14 @@ void QQuickMenuPrivate::itemSiblingOrderChanged(QQuickItem *)
     // reorder the restacked items (eg. by a Repeater)
     Q_Q(QQuickMenu);
     QList<QQuickItem *> siblings = contentItem->childItems();
+
+    int to = 0;
     for (int i = 0; i < siblings.count(); ++i) {
         QQuickItem* sibling = siblings.at(i);
+        if (QQuickItemPrivate::get(sibling)->isTransparentForPositioner())
+            continue;
         int index = contentModel->indexOf(sibling, nullptr);
-        q->moveItem(index, i);
+        q->moveItem(index, to++);
     }
 }
 
@@ -249,6 +254,31 @@ void QQuickMenuPrivate::setCurrentIndex(int index)
     contentItem->setProperty("currentIndex", index);
 }
 
+void QQuickMenuPrivate::activateNextItem()
+{
+    int index = currentIndex();
+    int count = contentModel->count();
+    while (++index < count) {
+        QQuickItem *item = itemAt(index);
+        if (!item || !item->activeFocusOnTab())
+            continue;
+        item->forceActiveFocus(Qt::TabFocusReason);
+        break;
+    }
+}
+
+void QQuickMenuPrivate::activatePreviousItem()
+{
+    int index = currentIndex();
+    while (--index >= 0) {
+        QQuickItem *item = itemAt(index);
+        if (!item || !item->activeFocusOnTab())
+            continue;
+        item->forceActiveFocus(Qt::BacktabFocusReason);
+        break;
+    }
+}
+
 void QQuickMenuPrivate::contentData_append(QQmlListProperty<QObject> *prop, QObject *obj)
 {
     QQuickMenuPrivate *p = static_cast<QQuickMenuPrivate *>(prop->data);
@@ -284,8 +314,8 @@ void QQuickMenuPrivate::contentData_clear(QQmlListProperty<QObject> *prop)
     p->contentData.clear();
 }
 
-QQuickMenu::QQuickMenu(QObject *parent) :
-    QQuickPopup(*(new QQuickMenuPrivate), parent)
+QQuickMenu::QQuickMenu(QObject *parent)
+    : QQuickPopup(*(new QQuickMenuPrivate), parent)
 {
     setFocus(true);
     setClosePolicy(CloseOnEscape | CloseOnPressOutside | CloseOnReleaseOutside);
@@ -461,8 +491,13 @@ void QQuickMenu::componentComplete()
 void QQuickMenu::contentItemChange(QQuickItem *newItem, QQuickItem *oldItem)
 {
     Q_D(QQuickMenu);
-    Q_UNUSED(oldItem);
     QQuickPopup::contentItemChange(newItem, oldItem);
+
+    if (oldItem)
+        QQuickItemPrivate::get(oldItem)->removeItemChangeListener(d, QQuickItemPrivate::Children);
+    if (newItem)
+        QQuickItemPrivate::get(newItem)->addItemChangeListener(d, QQuickItemPrivate::Children);
+
     d->contentItem = newItem;
 }
 
@@ -501,31 +536,29 @@ void QQuickMenu::keyReleaseEvent(QKeyEvent *event)
     // shown at once.
     switch (event->key()) {
     case Qt::Key_Up:
-        if (d->contentItem->metaObject()->indexOfMethod("decrementCurrentIndex()") != -1)
-            QMetaObject::invokeMethod(d->contentItem, "decrementCurrentIndex");
+        d->activatePreviousItem();
         break;
 
     case Qt::Key_Down:
-        if (d->contentItem->metaObject()->indexOfMethod("incrementCurrentIndex()") != -1)
-            QMetaObject::invokeMethod(d->contentItem, "incrementCurrentIndex");
+        d->activateNextItem();
         break;
 
     default:
         break;
     }
-
-    int index = d->currentIndex();
-    QQuickItem *item = itemAt(index);
-    if (item)
-        item->forceActiveFocus();
 }
 
-#ifndef QT_NO_ACCESSIBILITY
+QFont QQuickMenu::defaultFont() const
+{
+    return QQuickControlPrivate::themeFont(QPlatformTheme::MenuFont);
+}
+
+#if QT_CONFIG(accessibility)
 QAccessible::Role QQuickMenu::accessibleRole() const
 {
     return QAccessible::PopupMenu;
 }
-#endif // QT_NO_ACCESSIBILITY
+#endif
 
 QT_END_NAMESPACE
 
