@@ -1,6 +1,6 @@
 /****************************************************************************
 **
-** Copyright (C) 2016 The Qt Company Ltd.
+** Copyright (C) 2017 The Qt Company Ltd.
 ** Contact: http://www.qt.io/licensing/
 **
 ** This file is part of the Qt Quick Templates 2 module of the Qt Toolkit.
@@ -36,6 +36,7 @@
 
 #include "qquickoverlay_p.h"
 #include "qquickoverlay_p_p.h"
+#include "qquickpopupitem_p_p.h"
 #include "qquickpopup_p_p.h"
 #include "qquickdrawer_p_p.h"
 #include "qquickapplicationwindow_p.h"
@@ -79,7 +80,7 @@ static QQuickItem *createDimmer(QQmlComponent *component, QQuickPopup *popup, QQ
         QQmlContext *creationContext = component->creationContext();
         if (!creationContext)
             creationContext = qmlContext(parent);
-        QQmlContext *context = new QQmlContext(creationContext);
+        QQmlContext *context = new QQmlContext(creationContext, parent);
         context->setContextObject(popup);
         item = qobject_cast<QQuickItem*>(component->beginCreate(context));
     }
@@ -96,10 +97,15 @@ static QQuickItem *createDimmer(QQmlComponent *component, QQuickPopup *popup, QQ
         item->setZ(popup->z());
         if (popup->isModal()) {
             item->setAcceptedMouseButtons(Qt::AllButtons);
+#if QT_CONFIG(cursor)
+            item->setCursor(Qt::ArrowCursor);
+#endif
+#if QT_CONFIG(quicktemplates2_hover)
             // TODO: switch to QStyleHints::useHoverEffects in Qt 5.8
             item->setAcceptHoverEvents(true);
             // item->setAcceptHoverEvents(QGuiApplication::styleHints()->useHoverEffects());
             // connect(QGuiApplication::styleHints(), &QStyleHints::useHoverEffectsChanged, item, &QQuickItem::setAcceptHoverEvents);
+#endif
         }
         if (component)
             component->completeCreate();
@@ -163,18 +169,149 @@ QVector<QQuickDrawer *> QQuickOverlayPrivate::stackingOrderDrawers() const
     return sorted;
 }
 
-void QQuickOverlayPrivate::itemGeometryChanged(QQuickItem *, const QRectF &newGeometry, const QRectF &)
+void QQuickOverlayPrivate::itemGeometryChanged(QQuickItem *item, QQuickGeometryChange, const QRectF &)
 {
     Q_Q(QQuickOverlay);
-    q->setSize(newGeometry.size());
+    q->setSize(QSizeF(item->width(), item->height()));
 }
 
-QQuickOverlayPrivate::QQuickOverlayPrivate() :
-    modal(nullptr),
-    modeless(nullptr)
+QQuickOverlayPrivate::QQuickOverlayPrivate()
+    : modal(nullptr),
+      modeless(nullptr)
 {
 }
 
+bool QQuickOverlayPrivate::startDrag(QEvent *event, const QPointF &pos)
+{
+    Q_Q(QQuickOverlay);
+    if (allDrawers.isEmpty())
+        return false;
+
+    // don't start dragging a drawer if a modal popup overlay is blocking (QTBUG-60602)
+    QQuickItem *item = q->childAt(pos.x(), pos.y());
+    if (item) {
+        const auto popups = stackingOrderPopups();
+        for (QQuickPopup *popup : popups) {
+            QQuickPopupPrivate *p = QQuickPopupPrivate::get(popup);
+            if (p->dimmer == item && popup->isVisible() && popup->isModal())
+                return false;
+        }
+    }
+
+    const QVector<QQuickDrawer *> drawers = stackingOrderDrawers();
+    for (QQuickDrawer *drawer : drawers) {
+        QQuickDrawerPrivate *p = QQuickDrawerPrivate::get(drawer);
+        if (p->startDrag(event)) {
+            setMouseGrabberPopup(drawer);
+            return true;
+        }
+    }
+
+    return false;
+}
+
+bool QQuickOverlayPrivate::handlePress(QQuickItem *source, QEvent *event, QQuickPopup *target)
+{
+    if (target) {
+        if (target->overlayEvent(source, event)) {
+            setMouseGrabberPopup(target);
+            return true;
+        }
+        return false;
+    } else if (!mouseGrabberPopup) {
+        // allow non-modal popups to close themselves,
+        // and non-dimming modal popups to block the event
+        const auto popups = stackingOrderPopups();
+        for (QQuickPopup *popup : popups) {
+            if (popup->overlayEvent(source, event)) {
+                setMouseGrabberPopup(popup);
+                return true;
+            }
+        }
+    }
+
+    event->ignore();
+    return false;
+}
+
+bool QQuickOverlayPrivate::handleMove(QQuickItem *source, QEvent *event, QQuickPopup *target)
+{
+    if (target)
+        return target->overlayEvent(source, event);
+    return false;
+}
+
+bool QQuickOverlayPrivate::handleRelease(QQuickItem *source, QEvent *event, QQuickPopup *target)
+{
+    if (target) {
+        setMouseGrabberPopup(nullptr);
+        if (target->overlayEvent(source, event)) {
+            setMouseGrabberPopup(nullptr);
+            return true;
+        }
+    } else {
+        const auto popups = stackingOrderPopups();
+        for (QQuickPopup *popup : popups) {
+            if (popup->overlayEvent(source, event))
+                return true;
+        }
+    }
+    return false;
+}
+
+bool QQuickOverlayPrivate::handleMouseEvent(QQuickItem *source, QMouseEvent *event, QQuickPopup *target)
+{
+    switch (event->type()) {
+    case QEvent::MouseButtonPress:
+        if (!target && startDrag(event, event->windowPos()))
+            return true;
+        return handlePress(source, event, target);
+    case QEvent::MouseMove:
+        return handleMove(source, event, target ? target : mouseGrabberPopup.data());
+    case QEvent::MouseButtonRelease:
+        return handleRelease(source, event, target ? target : mouseGrabberPopup.data());
+    default:
+        break;
+    }
+    return false;
+}
+
+#if QT_CONFIG(quicktemplates2_multitouch)
+bool QQuickOverlayPrivate::handleTouchEvent(QQuickItem *source, QTouchEvent *event, QQuickPopup *target)
+{
+    bool handled = false;
+    switch (event->type()) {
+    case QEvent::TouchBegin:
+    case QEvent::TouchUpdate:
+    case QEvent::TouchEnd:
+        for (const QTouchEvent::TouchPoint &point : event->touchPoints()) {
+            switch (point.state()) {
+            case Qt::TouchPointPressed:
+                if (!target && startDrag(event, point.scenePos()))
+                    handled = true;
+                else
+                    handled |= handlePress(source, event, target);
+                break;
+            case Qt::TouchPointMoved:
+                handled |= handleMove(source, event, target ? target : mouseGrabberPopup.data());
+                break;
+            case Qt::TouchPointReleased:
+                handled |= handleRelease(source, event, target ? target : mouseGrabberPopup.data());
+                break;
+            default:
+                break;
+            }
+        }
+        break;
+
+    default:
+        break;
+    }
+
+    return handled;
+}
+#endif
+
 void QQuickOverlayPrivate::addPopup(QQuickPopup *popup)
 {
     Q_Q(QQuickOverlay);
@@ -212,6 +349,8 @@ QQuickOverlay::QQuickOverlay(QQuickItem *parent)
     if (parent) {
         setSize(QSizeF(parent->width(), parent->height()));
         QQuickItemPrivate::get(parent)->addItemChangeListener(d, QQuickItemPrivate::Geometry);
+        if (QQuickWindow *window = parent->window())
+            window->installEventFilter(this);
     }
 }
 
@@ -323,59 +462,30 @@ void QQuickOverlay::geometryChanged(const QRectF &newGeometry, const QRectF &old
 void QQuickOverlay::mousePressEvent(QMouseEvent *event)
 {
     Q_D(QQuickOverlay);
-    emit pressed();
-
-    if (!d->allDrawers.isEmpty()) {
-        // the overlay background was pressed, so there are no modal popups open.
-        // test if the press point lands on any drawer's drag margin
-
-        const QVector<QQuickDrawer *> drawers = d->stackingOrderDrawers();
-        for (QQuickDrawer *drawer : drawers) {
-            QQuickDrawerPrivate *p = QQuickDrawerPrivate::get(drawer);
-            if (p->startDrag(window(), event)) {
-                d->setMouseGrabberPopup(drawer);
-                return;
-            }
-        }
-    }
-
-    if (!d->mouseGrabberPopup) {
-        const auto popups = d->stackingOrderPopups();
-        for (QQuickPopup *popup : popups) {
-            if (popup->overlayEvent(this, event)) {
-                d->setMouseGrabberPopup(popup);
-                return;
-            }
-        }
-    }
-
-    event->ignore();
+    d->handleMouseEvent(this, event);
 }
 
 void QQuickOverlay::mouseMoveEvent(QMouseEvent *event)
 {
     Q_D(QQuickOverlay);
-    if (d->mouseGrabberPopup)
-        d->mouseGrabberPopup->overlayEvent(this, event);
+    d->handleMouseEvent(this, event);
 }
 
 void QQuickOverlay::mouseReleaseEvent(QMouseEvent *event)
 {
     Q_D(QQuickOverlay);
-    emit released();
-
-    if (d->mouseGrabberPopup) {
-        d->mouseGrabberPopup->overlayEvent(this, event);
-        d->setMouseGrabberPopup(nullptr);
-    } else {
-        const auto popups = d->stackingOrderPopups();
-        for (QQuickPopup *popup : popups) {
-            if (popup->overlayEvent(this, event))
-                break;
-        }
-    }
+    d->handleMouseEvent(this, event);
 }
 
+#if QT_CONFIG(quicktemplates2_multitouch)
+void QQuickOverlay::touchEvent(QTouchEvent *event)
+{
+    Q_D(QQuickOverlay);
+    d->handleTouchEvent(this, event);
+}
+#endif
+
+#if QT_CONFIG(wheelevent)
 void QQuickOverlay::wheelEvent(QWheelEvent *event)
 {
     Q_D(QQuickOverlay);
@@ -391,6 +501,7 @@ void QQuickOverlay::wheelEvent(QWheelEvent *event)
     }
     event->ignore();
 }
+#endif
 
 bool QQuickOverlay::childMouseEventFilter(QQuickItem *item, QEvent *event)
 {
@@ -409,19 +520,18 @@ bool QQuickOverlay::childMouseEventFilter(QQuickItem *item, QEvent *event)
         // does not have background dimming.
         if (item == p->dimmer || !p->popupItem->isAncestorOf(item)) {
             switch (event->type()) {
+#if QT_CONFIG(quicktemplates2_multitouch)
+            case QEvent::TouchBegin:
+            case QEvent::TouchUpdate:
+            case QEvent::TouchEnd:
+                return d->handleTouchEvent(item, static_cast<QTouchEvent *>(event), popup);
+#endif
+
             case QEvent::MouseButtonPress:
-                emit pressed();
-                if (popup->overlayEvent(item, event)) {
-                    d->setMouseGrabberPopup(popup);
-                    return true;
-                }
-                break;
             case QEvent::MouseMove:
-                return popup->overlayEvent(item, event);
             case QEvent::MouseButtonRelease:
-                emit released();
-                d->setMouseGrabberPopup(nullptr);
-                return popup->overlayEvent(item, event);
+                return d->handleMouseEvent(item, static_cast<QMouseEvent *>(event), popup);
+
             default:
                 break;
             }
@@ -430,4 +540,75 @@ bool QQuickOverlay::childMouseEventFilter(QQuickItem *item, QEvent *event)
     return false;
 }
 
+bool QQuickOverlay::eventFilter(QObject *object, QEvent *event)
+{
+    Q_D(QQuickOverlay);
+    if (!isVisible() || object != d->window)
+        return false;
+
+    switch (event->type()) {
+#if QT_CONFIG(quicktemplates2_multitouch)
+    case QEvent::TouchBegin:
+    case QEvent::TouchUpdate:
+    case QEvent::TouchEnd:
+        if (static_cast<QTouchEvent *>(event)->touchPointStates() & Qt::TouchPointPressed)
+            emit pressed();
+        if (static_cast<QTouchEvent *>(event)->touchPointStates() & Qt::TouchPointReleased)
+            emit released();
+
+        // allow non-modal popups to close on touch release outside
+        if (!d->mouseGrabberPopup) {
+            for (const QTouchEvent::TouchPoint &point : static_cast<QTouchEvent *>(event)->touchPoints()) {
+                if (point.state() == Qt::TouchPointReleased) {
+                    if (d->handleRelease(d->window->contentItem(), event, nullptr))
+                        break;
+                }
+            }
+        }
+
+        QQuickWindowPrivate::get(d->window)->handleTouchEvent(static_cast<QTouchEvent *>(event));
+
+        // If a touch event hasn't been accepted after being delivered, there
+        // were no items interested in touch events at any of the touch points.
+        // Make sure to accept the touch event in order to receive the consequent
+        // touch events, to be able to close non-modal popups on release outside.
+        event->accept();
+        return true;
+#endif
+
+    case QEvent::MouseButtonPress:
+#if QT_CONFIG(quicktemplates2_multitouch)
+        // do not emit pressed() twice when mouse events have been synthesized from touch events
+        if (static_cast<QMouseEvent *>(event)->source() == Qt::MouseEventNotSynthesized)
+#endif
+            emit pressed();
+
+        QQuickWindowPrivate::get(d->window)->handleMouseEvent(static_cast<QMouseEvent *>(event));
+
+        // If a mouse event hasn't been accepted after being delivered, there
+        // was no item interested in mouse events at the mouse point. Make sure
+        // to accept the mouse event in order to receive the consequent mouse
+        // events, to be able to close non-modal popups on release outside.
+        event->accept();
+        return true;
+
+    case QEvent::MouseButtonRelease:
+#if QT_CONFIG(quicktemplates2_multitouch)
+        // do not emit released() twice when mouse events have been synthesized from touch events
+        if (static_cast<QMouseEvent *>(event)->source() == Qt::MouseEventNotSynthesized)
+#endif
+            emit released();
+
+        // allow non-modal popups to close on mouse release outside
+        if (!d->mouseGrabberPopup)
+            d->handleRelease(d->window->contentItem(), event, nullptr);
+        break;
+
+    default:
+        break;
+    }
+
+    return false;
+}
+
 QT_END_NAMESPACE
