@@ -1,6 +1,6 @@
 /****************************************************************************
 **
-** Copyright (C) 2016 The Qt Company Ltd.
+** Copyright (C) 2017 The Qt Company Ltd.
 ** Contact: http://www.qt.io/licensing/
 **
 ** This file is part of the Qt Quick Templates 2 module of the Qt Toolkit.
@@ -49,20 +49,34 @@ QT_BEGIN_NAMESPACE
     \inqmlmodule QtQuick.Controls
     \since 5.7
     \ingroup qtquickcontrols2-input
-    \brief A slider control used to select a range of values.
+    \brief Used to select a range of values by sliding two handles along a track.
 
     \image qtquickcontrols2-rangeslider.gif
 
     RangeSlider is used to select a range specified by two values, by sliding
     each handle along a track.
 
+    In the example below, custom \l from and \l to values are set, and the
+    initial positions of the \l first and \l second handles are set:
+
     \code
     RangeSlider {
-        first.value: 0.25
-        second.value: 0.75
+        from: 1
+        to: 100
+        first.value: 25
+        second.value: 75
     }
     \endcode
 
+    The \l {first.position} and \l {second.position} properties are expressed as
+    fractions of the control's size, in the range \c {0.0 - 1.0}.
+    The \l {first.visualPosition} and \l {second.visualPosition} properties are
+    the same, except that they are reversed in a
+    \l {Right-to-left User Interfaces}{right-to-left} application.
+    The \c visualPosition is useful for positioning the handles when styling
+    RangeSlider. In the example above, \l {first.visualPosition} will be \c 0.24
+    in a left-to-right application, and \c 0.76 in a right-to-left application.
+
     \sa {Customizing RangeSlider}, {Input Controls}
 */
 
@@ -70,14 +84,16 @@ class QQuickRangeSliderNodePrivate : public QObjectPrivate
 {
     Q_DECLARE_PUBLIC(QQuickRangeSliderNode)
 public:
-    QQuickRangeSliderNodePrivate(qreal value, QQuickRangeSlider *slider) :
-        value(value),
-        isPendingValue(false),
-        pendingValue(0),
-        position(0),
-        handle(nullptr),
-        slider(slider),
-        pressed(false)
+    QQuickRangeSliderNodePrivate(qreal value, QQuickRangeSlider *slider)
+        : value(value),
+          isPendingValue(false),
+          pendingValue(0),
+          position(0),
+          handle(nullptr),
+          slider(slider),
+          pressed(false),
+          hovered(false),
+          touchId(-1)
     {
     }
 
@@ -88,9 +104,6 @@ public:
 
     static QQuickRangeSliderNodePrivate *get(QQuickRangeSliderNode *node);
 
-private:
-    friend class QQuickRangeSlider;
-
     qreal value;
     bool isPendingValue;
     qreal pendingValue;
@@ -98,6 +111,8 @@ private:
     QQuickItem *handle;
     QQuickRangeSlider *slider;
     bool pressed;
+    bool hovered;
+    int touchId;
 };
 
 bool QQuickRangeSliderNodePrivate::isFirst() const
@@ -132,8 +147,8 @@ QQuickRangeSliderNodePrivate *QQuickRangeSliderNodePrivate::get(QQuickRangeSlide
     return node->d_func();
 }
 
-QQuickRangeSliderNode::QQuickRangeSliderNode(qreal value, QQuickRangeSlider *slider) :
-    QObject(*(new QQuickRangeSliderNodePrivate(value, slider)), slider)
+QQuickRangeSliderNode::QQuickRangeSliderNode(qreal value, QQuickRangeSlider *slider)
+    : QObject(*(new QQuickRangeSliderNodePrivate(value, slider)), slider)
 {
 }
 
@@ -215,7 +230,7 @@ void QQuickRangeSliderNode::setHandle(QQuickItem *handle)
     if (d->handle == handle)
         return;
 
-    QQuickControlPrivate::get(d->slider)->deleteDelegate(d->handle);
+    QQuickControlPrivate::destroyDelegate(d->handle, d->slider);
     d->handle = handle;
     if (handle) {
         if (!handle->parentItem())
@@ -261,6 +276,22 @@ void QQuickRangeSliderNode::setPressed(bool pressed)
     emit pressedChanged();
 }
 
+bool QQuickRangeSliderNode::isHovered() const
+{
+    Q_D(const QQuickRangeSliderNode);
+    return d->hovered;
+}
+
+void QQuickRangeSliderNode::setHovered(bool hovered)
+{
+    Q_D(QQuickRangeSliderNode);
+    if (d->hovered == hovered)
+        return;
+
+    d->hovered = hovered;
+    emit hoveredChanged();
+}
+
 void QQuickRangeSliderNode::increase()
 {
     Q_D(QQuickRangeSliderNode);
@@ -283,23 +314,37 @@ class QQuickRangeSliderPrivate : public QQuickControlPrivate
     Q_DECLARE_PUBLIC(QQuickRangeSlider)
 
 public:
-    QQuickRangeSliderPrivate() :
-        from(defaultFrom),
-        to(defaultTo),
-        stepSize(0),
-        first(nullptr),
-        second(nullptr),
-        orientation(Qt::Horizontal),
-        snapMode(QQuickRangeSlider::NoSnap)
+    QQuickRangeSliderPrivate()
+        : live(true),
+          from(defaultFrom),
+          to(defaultTo),
+          stepSize(0),
+          first(nullptr),
+          second(nullptr),
+          orientation(Qt::Horizontal),
+          snapMode(QQuickRangeSlider::NoSnap)
     {
     }
 
+    QQuickRangeSliderNode *pressedNode(int touchId = -1) const;
+
+#if QT_CONFIG(quicktemplates2_multitouch)
+    bool acceptTouch(const QTouchEvent::TouchPoint &point) override;
+#endif
+    void handlePress(const QPointF &point) override;
+    void handleMove(const QPointF &point) override;
+    void handleRelease(const QPointF &point) override;
+    void handleUngrab() override;
+
+    void updateHover(const QPointF &pos);
+
+    bool live;
     qreal from;
     qreal to;
     qreal stepSize;
     QQuickRangeSliderNode *first;
     QQuickRangeSliderNode *second;
-    QPoint pressPoint;
+    QPointF pressPoint;
     Qt::Orientation orientation;
     QQuickRangeSlider::SnapMode snapMode;
 };
@@ -322,7 +367,7 @@ static qreal snapPosition(const QQuickRangeSlider *slider, qreal position)
     return qRound(position / effectiveStep) * effectiveStep;
 }
 
-static qreal positionAt(const QQuickRangeSlider *slider, QQuickItem *handle, const QPoint &point)
+static qreal positionAt(const QQuickRangeSlider *slider, QQuickItem *handle, const QPointF &point)
 {
     if (slider->orientation() == Qt::Horizontal) {
         const qreal hw = handle ? handle->width() : 0;
@@ -343,15 +388,167 @@ static qreal positionAt(const QQuickRangeSlider *slider, QQuickItem *handle, con
     return 0;
 }
 
-QQuickRangeSlider::QQuickRangeSlider(QQuickItem *parent) :
-    QQuickControl(*(new QQuickRangeSliderPrivate), parent)
+QQuickRangeSliderNode *QQuickRangeSliderPrivate::pressedNode(int touchId) const
+{
+    if (touchId == -1)
+        return first->isPressed() ? first : (second->isPressed() ? second : nullptr);
+    if (QQuickRangeSliderNodePrivate::get(first)->touchId == touchId)
+        return first;
+    if (QQuickRangeSliderNodePrivate::get(second)->touchId == touchId)
+        return second;
+    return nullptr;
+}
+
+#if QT_CONFIG(quicktemplates2_multitouch)
+bool QQuickRangeSliderPrivate::acceptTouch(const QTouchEvent::TouchPoint &point)
+{
+    int firstId = QQuickRangeSliderNodePrivate::get(first)->touchId;
+    int secondId = QQuickRangeSliderNodePrivate::get(second)->touchId;
+
+    if (((firstId == -1 || secondId == -1) && point.state() == Qt::TouchPointPressed) || point.id() == firstId || point.id() == secondId) {
+        touchId = point.id();
+        return true;
+    }
+
+    return false;
+}
+#endif
+
+void QQuickRangeSliderPrivate::handlePress(const QPointF &point)
+{
+    Q_Q(QQuickRangeSlider);
+    QQuickControlPrivate::handlePress(point);
+    pressPoint = point;
+
+    QQuickItem *firstHandle = first->handle();
+    QQuickItem *secondHandle = second->handle();
+    const bool firstHit = firstHandle && !first->isPressed() && firstHandle->contains(q->mapToItem(firstHandle, point));
+    const bool secondHit = secondHandle && !second->isPressed() && secondHandle->contains(q->mapToItem(secondHandle, point));
+    QQuickRangeSliderNode *hitNode = nullptr;
+    QQuickRangeSliderNode *otherNode = nullptr;
+
+    if (firstHit && secondHit) {
+        // choose highest
+        hitNode = firstHandle->z() > secondHandle->z() ? first : second;
+        otherNode = firstHandle->z() > secondHandle->z() ? second : first;
+    } else if (firstHit) {
+        hitNode = first;
+        otherNode = second;
+    } else if (secondHit) {
+        hitNode = second;
+        otherNode = first;
+    } else {
+        // find the nearest
+        const qreal firstPos = positionAt(q, firstHandle, point);
+        const qreal secondPos = positionAt(q, secondHandle, point);
+        const qreal firstDistance = qAbs(firstPos - first->position());
+        const qreal secondDistance = qAbs(secondPos - second->position());
+
+        if (qFuzzyCompare(firstDistance, secondDistance)) {
+            // same distance => choose the one that can be moved towards the press position
+            const bool inverted = from > to;
+            if ((!inverted && firstPos < first->position()) || (inverted && firstPos > first->position())) {
+                hitNode = first;
+                otherNode = second;
+            } else {
+                hitNode = second;
+                otherNode = first;
+            }
+        } else if (firstDistance < secondDistance) {
+            hitNode = first;
+            otherNode = second;
+        } else {
+            hitNode = second;
+            otherNode = first;
+        }
+    }
+
+    if (hitNode) {
+        hitNode->setPressed(true);
+        if (QQuickItem *handle = hitNode->handle())
+            handle->setZ(1);
+        QQuickRangeSliderNodePrivate::get(hitNode)->touchId = touchId;
+    }
+    if (otherNode) {
+        if (QQuickItem *handle = otherNode->handle())
+            handle->setZ(0);
+    }
+}
+
+void QQuickRangeSliderPrivate::handleMove(const QPointF &point)
+{
+    Q_Q(QQuickRangeSlider);
+    QQuickControlPrivate::handleMove(point);
+    QQuickRangeSliderNode *pressedNode = QQuickRangeSliderPrivate::pressedNode(touchId);
+    if (pressedNode) {
+        qreal pos = positionAt(q, pressedNode->handle(), point);
+        if (snapMode == QQuickRangeSlider::SnapAlways)
+            pos = snapPosition(q, pos);
+        if (live)
+            pressedNode->setValue(valueAt(q, pos));
+        else
+            QQuickRangeSliderNodePrivate::get(pressedNode)->setPosition(pos);
+    }
+}
+
+void QQuickRangeSliderPrivate::handleRelease(const QPointF &point)
+{
+    Q_Q(QQuickRangeSlider);
+    QQuickControlPrivate::handleRelease(point);
+    pressPoint = QPointF();
+
+    QQuickRangeSliderNode *pressedNode = QQuickRangeSliderPrivate::pressedNode(touchId);
+    if (!pressedNode)
+        return;
+    QQuickRangeSliderNodePrivate *pressedNodePrivate = QQuickRangeSliderNodePrivate::get(pressedNode);
+
+    if (q->keepMouseGrab() || q->keepTouchGrab()) {
+        qreal pos = positionAt(q, pressedNode->handle(), point);
+        if (snapMode != QQuickRangeSlider::NoSnap)
+            pos = snapPosition(q, pos);
+        qreal val = valueAt(q, pos);
+        if (!qFuzzyCompare(val, pressedNode->value()))
+            pressedNode->setValue(val);
+        else if (snapMode != QQuickRangeSlider::NoSnap)
+            pressedNodePrivate->setPosition(pos);
+        q->setKeepMouseGrab(false);
+        q->setKeepTouchGrab(false);
+    }
+    pressedNode->setPressed(false);
+    pressedNodePrivate->touchId = -1;
+}
+
+void QQuickRangeSliderPrivate::handleUngrab()
+{
+    QQuickControlPrivate::handleUngrab();
+    pressPoint = QPointF();
+    first->setPressed(false);
+    second->setPressed(false);
+    QQuickRangeSliderNodePrivate::get(first)->touchId = -1;
+    QQuickRangeSliderNodePrivate::get(second)->touchId = -1;
+}
+
+void QQuickRangeSliderPrivate::updateHover(const QPointF &pos)
+{
+    Q_Q(QQuickRangeSlider);
+    QQuickItem *firstHandle = first->handle();
+    QQuickItem *secondHandle = second->handle();
+    first->setHovered(firstHandle && firstHandle->isEnabled() && firstHandle->contains(q->mapToItem(firstHandle, pos)));
+    second->setHovered(secondHandle && secondHandle->isEnabled() && secondHandle->contains(q->mapToItem(secondHandle, pos)));
+}
+
+QQuickRangeSlider::QQuickRangeSlider(QQuickItem *parent)
+    : QQuickControl(*(new QQuickRangeSliderPrivate), parent)
 {
     Q_D(QQuickRangeSlider);
     d->first = new QQuickRangeSliderNode(0.0, this);
     d->second = new QQuickRangeSliderNode(1.0, this);
 
-    setAcceptedMouseButtons(Qt::LeftButton);
     setFlag(QQuickItem::ItemIsFocusScope);
+    setAcceptedMouseButtons(Qt::LeftButton);
+#if QT_CONFIG(cursor)
+    setCursor(Qt::ArrowCursor);
+#endif
 }
 
 /*!
@@ -417,6 +614,7 @@ void QQuickRangeSlider::setTo(qreal to)
     \qmlproperty real QtQuick.Controls::RangeSlider::first.visualPosition
     \qmlproperty Item QtQuick.Controls::RangeSlider::first.handle
     \qmlproperty bool QtQuick.Controls::RangeSlider::first.pressed
+    \qmlproperty bool QtQuick.Controls::RangeSlider::first.hovered
 
     \table
     \header
@@ -430,9 +628,6 @@ void QQuickRangeSlider::setTo(qreal to)
             If \l to is greater than \l from, the value of the first handle
             must be greater than the second, and vice versa.
 
-            Unlike \l {first.position}{position}, value is not updated while the
-            handle is dragged, but rather when it has been released.
-
             The default value is \c 0.0.
     \row
         \li handle
@@ -441,7 +636,7 @@ void QQuickRangeSlider::setTo(qreal to)
         \li visualPosition
         \li This property holds the visual position of the first handle.
 
-            The position is defined as a percentage of the control's size, scaled to
+            The position is expressed as a fraction of the control's size, in the range
             \c {0.0 - 1.0}. When the control is \l {Control::mirrored}{mirrored}, the
             value is equal to \c {1.0 - position}. This makes the value suitable for
             visualizing the slider, taking right-to-left support into account.
@@ -449,14 +644,16 @@ void QQuickRangeSlider::setTo(qreal to)
         \li position
         \li This property holds the logical position of the first handle.
 
-            The position is defined as a percentage of the control's size, scaled
-            to \c {0.0 - 1.0}. Unlike \l {first.value}{value}, position is
-            continuously updated while the handle is dragged. For visualizing a
-            slider, the right-to-left aware
+            The position is expressed as a fraction of the control's size, in the range
+            \c {0.0 - 1.0}. For visualizing a slider, the right-to-left aware
             \l {first.visualPosition}{visualPosition} should be used instead.
     \row
         \li pressed
         \li This property holds whether the first handle is pressed.
+    \row
+        \li hovered
+        \li This property holds whether the first handle is hovered.
+            This property was introduced in QtQuick.Controls 2.1.
     \endtable
 
     \sa first.increase(), first.decrease()
@@ -474,6 +671,7 @@ QQuickRangeSliderNode *QQuickRangeSlider::first() const
     \qmlproperty real QtQuick.Controls::RangeSlider::second.visualPosition
     \qmlproperty Item QtQuick.Controls::RangeSlider::second.handle
     \qmlproperty bool QtQuick.Controls::RangeSlider::second.pressed
+    \qmlproperty bool QtQuick.Controls::RangeSlider::second.hovered
 
     \table
     \header
@@ -487,9 +685,6 @@ QQuickRangeSliderNode *QQuickRangeSlider::first() const
             If \l to is greater than \l from, the value of the first handle
             must be greater than the second, and vice versa.
 
-            Unlike \l {second.position}{position}, value is not updated while the
-            handle is dragged, but rather when it has been released.
-
             The default value is \c 0.0.
     \row
         \li handle
@@ -498,7 +693,7 @@ QQuickRangeSliderNode *QQuickRangeSlider::first() const
         \li visualPosition
         \li This property holds the visual position of the second handle.
 
-            The position is defined as a percentage of the control's size, scaled to
+            The position is expressed as a fraction of the control's size, in the range
             \c {0.0 - 1.0}. When the control is \l {Control::mirrored}{mirrored}, the
             value is equal to \c {1.0 - position}. This makes the value suitable for
             visualizing the slider, taking right-to-left support into account.
@@ -506,14 +701,16 @@ QQuickRangeSliderNode *QQuickRangeSlider::first() const
         \li position
         \li This property holds the logical position of the second handle.
 
-            The position is defined as a percentage of the control's size, scaled
-            to \c {0.0 - 1.0}. Unlike \l {second.value}{value}, position is
-            continuously updated while the handle is dragged. For visualizing a
-            slider, the right-to-left aware
+            The position is expressed as a fraction of the control's size, in the range
+            \c {0.0 - 1.0}. For visualizing a slider, the right-to-left aware
             \l {second.visualPosition}{visualPosition} should be used instead.
     \row
         \li pressed
         \li This property holds whether the second handle is pressed.
+    \row
+        \li hovered
+        \li This property holds whether the second handle is hovered.
+            This property was introduced in QtQuick.Controls 2.1.
     \endtable
 
     \sa second.increase(), second.decrease()
@@ -557,6 +754,9 @@ void QQuickRangeSlider::setStepSize(qreal step)
     \value RangeSlider.SnapAlways The slider snaps while the handle is dragged.
     \value RangeSlider.SnapOnRelease The slider does not snap while being dragged, but only after the handle is released.
 
+    For visual explanations of the various modes, see the
+    \l {Slider::}{snapMode} documentation of \l Slider.
+
     \sa stepSize
 */
 QQuickRangeSlider::SnapMode QQuickRangeSlider::snapMode() const
@@ -600,6 +800,33 @@ void QQuickRangeSlider::setOrientation(Qt::Orientation orientation)
     emit orientationChanged();
 }
 
+/*!
+    \since QtQuick.Controls 2.2 (Qt 5.9)
+    \qmlproperty bool QtQuick.Controls::RangeSlider::live
+
+    This property holds whether the slider provides live updates for the \l first.value
+    and \l second.value properties while the respective handles are dragged.
+
+    The default value is \c true.
+
+    \sa first.value, second.value
+*/
+bool QQuickRangeSlider::live() const
+{
+    Q_D(const QQuickRangeSlider);
+    return d->live;
+}
+
+void QQuickRangeSlider::setLive(bool live)
+{
+    Q_D(QQuickRangeSlider);
+    if (d->live == live)
+        return;
+
+    d->live = live;
+    emit liveChanged();
+}
+
 /*!
     \qmlmethod void QtQuick.Controls::RangeSlider::setValues(real firstValue, real secondValue)
 
@@ -714,6 +941,28 @@ void QQuickRangeSlider::keyPressEvent(QKeyEvent *event)
     }
 }
 
+void QQuickRangeSlider::hoverEnterEvent(QHoverEvent *event)
+{
+    Q_D(QQuickRangeSlider);
+    QQuickControl::hoverEnterEvent(event);
+    d->updateHover(event->posF());
+}
+
+void QQuickRangeSlider::hoverMoveEvent(QHoverEvent *event)
+{
+    Q_D(QQuickRangeSlider);
+    QQuickControl::hoverMoveEvent(event);
+    d->updateHover(event->posF());
+}
+
+void QQuickRangeSlider::hoverLeaveEvent(QHoverEvent *event)
+{
+    Q_D(QQuickRangeSlider);
+    QQuickControl::hoverLeaveEvent(event);
+    d->first->setHovered(false);
+    d->second->setHovered(false);
+}
+
 void QQuickRangeSlider::keyReleaseEvent(QKeyEvent *event)
 {
     Q_D(QQuickRangeSlider);
@@ -726,114 +975,60 @@ void QQuickRangeSlider::mousePressEvent(QMouseEvent *event)
 {
     Q_D(QQuickRangeSlider);
     QQuickControl::mousePressEvent(event);
-    d->pressPoint = event->pos();
-
-    QQuickItem *firstHandle = d->first->handle();
-    QQuickItem *secondHandle = d->second->handle();
-    const bool firstHit = firstHandle && firstHandle->contains(mapToItem(firstHandle, d->pressPoint));
-    const bool secondHit = secondHandle && secondHandle->contains(mapToItem(secondHandle, d->pressPoint));
-    QQuickRangeSliderNode *hitNode = nullptr;
-    QQuickRangeSliderNode *otherNode = nullptr;
-
-    if (firstHit && secondHit) {
-        // choose highest
-        hitNode = firstHandle->z() > secondHandle->z() ? d->first : d->second;
-        otherNode = firstHandle->z() > secondHandle->z() ? d->second : d->first;
-    } else if (firstHit) {
-        hitNode = d->first;
-        otherNode = d->second;
-    } else if (secondHit) {
-        hitNode = d->second;
-        otherNode = d->first;
-    } else {
-        // find the nearest
-        const qreal firstDistance = QLineF(firstHandle->boundingRect().center(),
-                                           mapToItem(firstHandle, event->pos())).length();
-        const qreal secondDistance = QLineF(secondHandle->boundingRect().center(),
-                                            mapToItem(secondHandle, event->pos())).length();
-
-        if (qFuzzyCompare(firstDistance, secondDistance)) {
-            // same distance => choose the one that can be moved towards the press position
-            const bool inverted = d->from > d->to;
-            const qreal pos = positionAt(this, firstHandle, event->pos());
-            if ((!inverted && pos < d->first->position()) || (inverted && pos > d->first->position())) {
-                hitNode = d->first;
-                otherNode = d->second;
-            } else {
-                hitNode = d->second;
-                otherNode = d->first;
-            }
-        } else if (firstDistance < secondDistance) {
-            hitNode = d->first;
-            otherNode = d->second;
-        } else {
-            hitNode = d->second;
-            otherNode = d->first;
-        }
-    }
-
-    if (hitNode) {
-        hitNode->setPressed(true);
-        hitNode->handle()->setZ(1);
-    }
-    if (otherNode)
-        otherNode->handle()->setZ(0);
+    d->handleMove(event->localPos());
 }
 
 void QQuickRangeSlider::mouseMoveEvent(QMouseEvent *event)
 {
     Q_D(QQuickRangeSlider);
-    QQuickControl::mouseMoveEvent(event);
     if (!keepMouseGrab()) {
         if (d->orientation == Qt::Horizontal)
-            setKeepMouseGrab(QQuickWindowPrivate::dragOverThreshold(event->pos().x() - d->pressPoint.x(), Qt::XAxis, event));
+            setKeepMouseGrab(QQuickWindowPrivate::dragOverThreshold(event->localPos().x() - d->pressPoint.x(), Qt::XAxis, event));
         else
-            setKeepMouseGrab(QQuickWindowPrivate::dragOverThreshold(event->pos().y() - d->pressPoint.y(), Qt::YAxis, event));
+            setKeepMouseGrab(QQuickWindowPrivate::dragOverThreshold(event->localPos().y() - d->pressPoint.y(), Qt::YAxis, event));
     }
-    if (keepMouseGrab()) {
-        QQuickRangeSliderNode *pressedNode = d->first->isPressed() ? d->first : (d->second->isPressed() ? d->second : nullptr);
-        if (pressedNode) {
-            qreal pos = positionAt(this, pressedNode->handle(), event->pos());
-            if (d->snapMode == SnapAlways)
-                pos = snapPosition(this, pos);
-            QQuickRangeSliderNodePrivate::get(pressedNode)->setPosition(pos);
+    QQuickControl::mouseMoveEvent(event);
+}
+
+#if QT_CONFIG(quicktemplates2_multitouch)
+void QQuickRangeSlider::touchEvent(QTouchEvent *event)
+{
+    Q_D(QQuickRangeSlider);
+    switch (event->type()) {
+    case QEvent::TouchUpdate:
+        for (const QTouchEvent::TouchPoint &point : event->touchPoints()) {
+            if (!d->acceptTouch(point))
+                continue;
+
+            switch (point.state()) {
+            case Qt::TouchPointPressed:
+                d->handlePress(point.pos());
+                break;
+            case Qt::TouchPointMoved:
+                if (!keepTouchGrab()) {
+                    if (d->orientation == Qt::Horizontal)
+                        setKeepTouchGrab(QQuickWindowPrivate::dragOverThreshold(point.pos().x() - point.startPos().x(), Qt::XAxis, &point));
+                    else
+                        setKeepTouchGrab(QQuickWindowPrivate::dragOverThreshold(point.pos().y() - point.startPos().y(), Qt::YAxis, &point));
+                }
+                if (keepTouchGrab())
+                    d->handleMove(point.pos());
+                break;
+            case Qt::TouchPointReleased:
+                d->handleRelease(point.pos());
+                break;
+            default:
+                break;
+            }
         }
+        break;
+
+    default:
+        QQuickControl::touchEvent(event);
+        break;
     }
 }
-
-void QQuickRangeSlider::mouseReleaseEvent(QMouseEvent *event)
-{
-    Q_D(QQuickRangeSlider);
-    QQuickControl::mouseReleaseEvent(event);
-
-    d->pressPoint = QPoint();
-    if (!keepMouseGrab())
-        return;
-
-    QQuickRangeSliderNode *pressedNode = d->first->isPressed() ? d->first : (d->second->isPressed() ? d->second : nullptr);
-    if (!pressedNode)
-        return;
-
-    qreal pos = positionAt(this, pressedNode->handle(), event->pos());
-    if (d->snapMode != NoSnap)
-        pos = snapPosition(this, pos);
-    qreal val = valueAt(this, pos);
-    if (!qFuzzyCompare(val, pressedNode->value()))
-        pressedNode->setValue(val);
-    else if (d->snapMode != NoSnap)
-        QQuickRangeSliderNodePrivate::get(pressedNode)->setPosition(pos);
-    setKeepMouseGrab(false);
-    pressedNode->setPressed(false);
-}
-
-void QQuickRangeSlider::mouseUngrabEvent()
-{
-    Q_D(QQuickRangeSlider);
-    QQuickControl::mouseUngrabEvent();
-    d->pressPoint = QPoint();
-    d->first->setPressed(false);
-    d->second->setPressed(false);
-}
+#endif
 
 void QQuickRangeSlider::mirrorChange()
 {
@@ -905,7 +1100,7 @@ void QQuickRangeSlider::componentComplete()
     \sa second
 */
 
-#ifndef QT_NO_ACCESSIBILITY
+#if QT_CONFIG(accessibility)
 QAccessible::Role QQuickRangeSlider::accessibleRole() const
 {
     return QAccessible::Slider;
