@@ -1,6 +1,6 @@
 /****************************************************************************
 **
-** Copyright (C) 2016 The Qt Company Ltd.
+** Copyright (C) 2017 The Qt Company Ltd.
 ** Contact: http://www.qt.io/licensing/
 **
 ** This file is part of the Qt Quick Templates 2 module of the Qt Toolkit.
@@ -43,6 +43,7 @@
 #include <QtQml/qqmlinfo.h>
 #include <QtQml/private/qqmllocale_p.h>
 #include <QtQml/private/qqmlengine_p.h>
+#include <QtQuick/private/qquicktextinput_p.h>
 
 QT_BEGIN_NAMESPACE
 
@@ -57,7 +58,7 @@ static const int AUTO_REPEAT_INTERVAL = 100;
     \inqmlmodule QtQuick.Controls
     \since 5.7
     \ingroup input
-    \brief A spinbox control that allows the user to select from a set of preset values.
+    \brief Allows the user to select from a set of preset values.
 
     \image qtquickcontrols2-spinbox.png
 
@@ -92,16 +93,39 @@ static const int AUTO_REPEAT_INTERVAL = 100;
     \sa Tumbler, {Customizing SpinBox}
 */
 
+/*!
+    \since QtQuick.Controls 2.2 (Qt 5.9)
+    \qmlsignal QtQuick.Controls::SpinBox::valueModified()
+
+    This signal is emitted when the spin box value has been interactively
+    modified by the user by either touch, mouse, wheel, or keys.
+*/
+
 class QQuickSpinBoxPrivate : public QQuickControlPrivate
 {
     Q_DECLARE_PUBLIC(QQuickSpinBox)
 
 public:
-    QQuickSpinBoxPrivate() : editable(false), from(0), to(99), value(0), stepSize(1),
-        delayTimer(0), repeatTimer(0), up(nullptr), down(nullptr), validator(nullptr) { }
+    QQuickSpinBoxPrivate()
+        : editable(false),
+          from(0),
+          to(99),
+          value(0),
+          stepSize(1),
+          delayTimer(0),
+          repeatTimer(0),
+          up(nullptr),
+          down(nullptr),
+          validator(nullptr),
+          inputMethodHints(Qt::ImhDigitsOnly)
+    {
+    }
 
     int boundValue(int value) const;
     void updateValue();
+    bool setValue(int value, bool modified);
+    void increase(bool modified);
+    void decrease(bool modified);
 
     int effectiveStepSize() const;
 
@@ -109,15 +133,16 @@ public:
     void updateUpEnabled();
     bool downEnabled() const;
     void updateDownEnabled();
+    void updateHover(const QPointF &pos);
 
     void startRepeatDelay();
     void startPressRepeat();
     void stopPressRepeat();
 
-    bool handleMousePressEvent(QQuickItem *child, QMouseEvent *event);
-    bool handleMouseMoveEvent(QQuickItem *child, QMouseEvent *event);
-    bool handleMouseReleaseEvent(QQuickItem *child, QMouseEvent *event);
-    bool handleMouseUngrabEvent(QQuickItem *child);
+    void handlePress(const QPointF &point) override;
+    void handleMove(const QPointF &point) override;
+    void handleRelease(const QPointF &point) override;
+    void handleUngrab() override;
 
     bool editable;
     int from;
@@ -131,6 +156,7 @@ public:
     QValidator *validator;
     mutable QJSValue textFromValue;
     mutable QJSValue valueFromText;
+    Qt::InputMethodHints inputMethodHints;
 };
 
 int QQuickSpinBoxPrivate::boundValue(int value) const
@@ -149,12 +175,42 @@ void QQuickSpinBoxPrivate::updateValue()
                 QV4::ExecutionEngine *v4 = QQmlEnginePrivate::getV4Engine(engine);
                 QJSValue loc(v4, QQmlLocale::wrap(v4, locale));
                 QJSValue val = q->valueFromText().call(QJSValueList() << text.toString() << loc);
-                q->setValue(val.toInt());
+                setValue(val.toInt(), true);
             }
         }
     }
 }
 
+bool QQuickSpinBoxPrivate::setValue(int newValue, bool modified)
+{
+    Q_Q(QQuickSpinBox);
+    if (q->isComponentComplete())
+        newValue = boundValue(newValue);
+
+    if (value == newValue)
+        return false;
+
+    value = newValue;
+
+    updateUpEnabled();
+    updateDownEnabled();
+
+    emit q->valueChanged();
+    if (modified)
+        emit q->valueModified();
+    return true;
+}
+
+void QQuickSpinBoxPrivate::increase(bool modified)
+{
+    setValue(value + effectiveStepSize(), modified);
+}
+
+void QQuickSpinBoxPrivate::decrease(bool modified)
+{
+    setValue(value - effectiveStepSize(), modified);
+}
+
 int QQuickSpinBoxPrivate::effectiveStepSize() const
 {
     return from > to ? -1 * stepSize : stepSize;
@@ -190,6 +246,15 @@ void QQuickSpinBoxPrivate::updateDownEnabled()
     downIndicator->setEnabled(from < to ? value > from : value < from);
 }
 
+void QQuickSpinBoxPrivate::updateHover(const QPointF &pos)
+{
+    Q_Q(QQuickSpinBox);
+    QQuickItem *ui = up->indicator();
+    QQuickItem *di = down->indicator();
+    up->setHovered(ui && ui->isEnabled() && ui->contains(q->mapToItem(ui, pos)));
+    down->setHovered(di && di->isEnabled() && di->contains(q->mapToItem(di, pos)));
+}
+
 void QQuickSpinBoxPrivate::startRepeatDelay()
 {
     Q_Q(QQuickSpinBox);
@@ -217,69 +282,73 @@ void QQuickSpinBoxPrivate::stopPressRepeat()
     }
 }
 
-bool QQuickSpinBoxPrivate::handleMousePressEvent(QQuickItem *child, QMouseEvent *event)
+void QQuickSpinBoxPrivate::handlePress(const QPointF &point)
 {
     Q_Q(QQuickSpinBox);
+    QQuickControlPrivate::handlePress(point);
     QQuickItem *ui = up->indicator();
     QQuickItem *di = down->indicator();
-    up->setPressed(ui && ui->isEnabled() && ui->contains(ui->mapFromItem(child, event->pos())));
-    down->setPressed(di && di->isEnabled() && di->contains(di->mapFromItem(child, event->pos())));
+    up->setPressed(ui && ui->isEnabled() && ui->contains(ui->mapFromItem(q, point)));
+    down->setPressed(di && di->isEnabled() && di->contains(di->mapFromItem(q, point)));
 
     bool pressed = up->isPressed() || down->isPressed();
     q->setAccessibleProperty("pressed", pressed);
     if (pressed)
         startRepeatDelay();
-    return pressed;
 }
 
-bool QQuickSpinBoxPrivate::handleMouseMoveEvent(QQuickItem *child, QMouseEvent *event)
+void QQuickSpinBoxPrivate::handleMove(const QPointF &point)
 {
     Q_Q(QQuickSpinBox);
+    QQuickControlPrivate::handleMove(point);
     QQuickItem *ui = up->indicator();
     QQuickItem *di = down->indicator();
-    up->setPressed(ui && ui->isEnabled() && ui->contains(ui->mapFromItem(child, event->pos())));
-    down->setPressed(di && di->isEnabled() && di->contains(di->mapFromItem(child, event->pos())));
+    up->setPressed(ui && ui->isEnabled() && ui->contains(ui->mapFromItem(q, point)));
+    down->setPressed(di && di->isEnabled() && di->contains(di->mapFromItem(q, point)));
 
     bool pressed = up->isPressed() || down->isPressed();
     q->setAccessibleProperty("pressed", pressed);
-    stopPressRepeat();
-    return pressed;
+    if (!pressed)
+        stopPressRepeat();
 }
 
-bool QQuickSpinBoxPrivate::handleMouseReleaseEvent(QQuickItem *child, QMouseEvent *event)
+void QQuickSpinBoxPrivate::handleRelease(const QPointF &point)
 {
     Q_Q(QQuickSpinBox);
+    QQuickControlPrivate::handleRelease(point);
     QQuickItem *ui = up->indicator();
     QQuickItem *di = down->indicator();
-    bool wasPressed = up->isPressed() || down->isPressed();
+
+    int oldValue = value;
     if (up->isPressed()) {
         up->setPressed(false);
-        if (repeatTimer <= 0 && ui && ui->contains(ui->mapFromItem(child, event->pos())))
+        if (repeatTimer <= 0 && ui && ui->contains(ui->mapFromItem(q, point)))
             q->increase();
     } else if (down->isPressed()) {
         down->setPressed(false);
-        if (repeatTimer <= 0 && di && di->contains(di->mapFromItem(child, event->pos())))
+        if (repeatTimer <= 0 && di && di->contains(di->mapFromItem(q, point)))
             q->decrease();
     }
+    if (value != oldValue)
+        emit q->valueModified();
 
     q->setAccessibleProperty("pressed", false);
     stopPressRepeat();
-    return wasPressed;
 }
 
-bool QQuickSpinBoxPrivate::handleMouseUngrabEvent(QQuickItem *)
+void QQuickSpinBoxPrivate::handleUngrab()
 {
     Q_Q(QQuickSpinBox);
+    QQuickControlPrivate::handleUngrab();
     up->setPressed(false);
     down->setPressed(false);
 
     q->setAccessibleProperty("pressed", false);
     stopPressRepeat();
-    return false;
 }
 
-QQuickSpinBox::QQuickSpinBox(QQuickItem *parent) :
-    QQuickControl(*(new QQuickSpinBoxPrivate), parent)
+QQuickSpinBox::QQuickSpinBox(QQuickItem *parent)
+    : QQuickControl(*(new QQuickSpinBoxPrivate), parent)
 {
     Q_D(QQuickSpinBox);
     d->up = new QQuickSpinButton(this);
@@ -288,6 +357,9 @@ QQuickSpinBox::QQuickSpinBox(QQuickItem *parent) :
     setFlag(ItemIsFocusScope);
     setFiltersChildMouseEvents(true);
     setAcceptedMouseButtons(Qt::LeftButton);
+#if QT_CONFIG(cursor)
+    setCursor(Qt::ArrowCursor);
+#endif
 }
 
 /*!
@@ -311,8 +383,12 @@ void QQuickSpinBox::setFrom(int from)
 
     d->from = from;
     emit fromChanged();
-    if (isComponentComplete())
-        setValue(d->value);
+    if (isComponentComplete()) {
+        if (!d->setValue(d->value, false)) {
+            d->updateUpEnabled();
+            d->updateDownEnabled();
+        }
+    }
 }
 
 /*!
@@ -336,8 +412,12 @@ void QQuickSpinBox::setTo(int to)
 
     d->to = to;
     emit toChanged();
-    if (isComponentComplete())
-        setValue(d->value);
+    if (isComponentComplete()) {
+        if (!d->setValue(d->value, false)) {
+            d->updateUpEnabled();
+            d->updateDownEnabled();
+        }
+    }
 }
 
 /*!
@@ -354,18 +434,7 @@ int QQuickSpinBox::value() const
 void QQuickSpinBox::setValue(int value)
 {
     Q_D(QQuickSpinBox);
-    if (isComponentComplete())
-        value = d->boundValue(value);
-
-    if (d->value == value)
-        return;
-
-    d->value = value;
-
-    d->updateUpEnabled();
-    d->updateDownEnabled();
-
-    emit valueChanged();
+    d->setValue(value, false);
 }
 
 /*!
@@ -410,7 +479,17 @@ void QQuickSpinBox::setEditable(bool editable)
     if (d->editable == editable)
         return;
 
+#if QT_CONFIG(cursor)
+    if (d->contentItem) {
+        if (editable)
+            d->contentItem->setCursor(Qt::IBeamCursor);
+        else
+            d->contentItem->unsetCursor();
+    }
+#endif
+
     d->editable = editable;
+    setAccessibleProperty("editable", editable);
     emit editableChanged();
 }
 
@@ -420,7 +499,16 @@ void QQuickSpinBox::setEditable(bool editable)
     This property holds the input text validator for editable spinboxes. By
     default, SpinBox uses \l IntValidator to accept input of integer numbers.
 
-    \snippet SpinBox.qml validator
+    \code
+    SpinBox {
+        id: control
+        validator: IntValidator {
+            locale: control.locale.name
+            bottom: Math.min(control.from, control.to)
+            top: Math.max(control.from, control.to)
+        }
+    }
+    \endcode
 
     \sa editable, textFromValue, valueFromText, {Control::locale}{locale}
 */
@@ -446,6 +534,12 @@ void QQuickSpinBox::setValidator(QValidator *validator)
     This property holds a callback function that is called whenever
     an integer value needs to be converted to display text.
 
+    The default function can be overridden to display custom text for a given
+    value. This applies to both editable and non-editable spinboxes;
+    for example, when using the up and down buttons or a mouse wheel to
+    increment and decrement the value, the new value is converted to display
+    text using this function.
+
     The callback function signature is \c {string function(value, locale)}.
     The function can have one or two arguments, where the first argument
     is the value to be converted, and the optional second argument is the
@@ -457,6 +551,10 @@ void QQuickSpinBox::setValidator(QValidator *validator)
     textFromValue: function(value, locale) { return Number(value).toLocaleString(locale, 'f', 0); }
     \endcode
 
+    \note When applying a custom \c textFromValue implementation for editable
+    spinboxes, a matching \l valueFromText implementation must be provided
+    to be able to convert the custom text back to an integer value.
+
     \sa valueFromText, validator, {Control::locale}{locale}
 */
 QJSValue QQuickSpinBox::textFromValue() const
@@ -474,7 +572,7 @@ void QQuickSpinBox::setTextFromValue(const QJSValue &callback)
 {
     Q_D(QQuickSpinBox);
     if (!callback.isCallable()) {
-        qmlInfo(this) << "textFromValue must be a callable function";
+        qmlWarning(this) << "textFromValue must be a callable function";
         return;
     }
     d->textFromValue = callback;
@@ -487,6 +585,9 @@ void QQuickSpinBox::setTextFromValue(const QJSValue &callback)
     This property holds a callback function that is called whenever
     input text needs to be converted to an integer value.
 
+    This function only needs to be overridden when \l textFromValue
+    is overridden for an editable spinbox.
+
     The callback function signature is \c {int function(text, locale)}.
     The function can have one or two arguments, where the first argument
     is the text to be converted, and the optional second argument is the
@@ -498,6 +599,10 @@ void QQuickSpinBox::setTextFromValue(const QJSValue &callback)
     valueFromText: function(text, locale) { return Number.fromLocaleString(locale, text); }
     \endcode
 
+    \note When applying a custom \l textFromValue implementation for editable
+    spinboxes, a matching \c valueFromText implementation must be provided
+    to be able to convert the custom text back to an integer value.
+
     \sa textFromValue, validator, {Control::locale}{locale}
 */
 QJSValue QQuickSpinBox::valueFromText() const
@@ -515,7 +620,7 @@ void QQuickSpinBox::setValueFromText(const QJSValue &callback)
 {
     Q_D(QQuickSpinBox);
     if (!callback.isCallable()) {
-        qmlInfo(this) << "valueFromText must be a callable function";
+        qmlWarning(this) << "valueFromText must be a callable function";
         return;
     }
     d->valueFromText = callback;
@@ -526,8 +631,10 @@ void QQuickSpinBox::setValueFromText(const QJSValue &callback)
     \qmlpropertygroup QtQuick.Controls::SpinBox::up
     \qmlproperty bool QtQuick.Controls::SpinBox::up.pressed
     \qmlproperty Item QtQuick.Controls::SpinBox::up.indicator
+    \qmlproperty bool QtQuick.Controls::SpinBox::up.hovered
 
-    These properties hold the up indicator item and whether it is pressed.
+    These properties hold the up indicator item and whether it is pressed or
+    hovered. The \c up.hovered property was introduced in QtQuick.Controls 2.1.
 
     \sa increase()
 */
@@ -541,8 +648,10 @@ QQuickSpinButton *QQuickSpinBox::up() const
     \qmlpropertygroup QtQuick.Controls::SpinBox::down
     \qmlproperty bool QtQuick.Controls::SpinBox::down.pressed
     \qmlproperty Item QtQuick.Controls::SpinBox::down.indicator
+    \qmlproperty bool QtQuick.Controls::SpinBox::down.hovered
 
-    These properties hold the down indicator item and whether it is pressed.
+    These properties hold the down indicator item and whether it is pressed or
+    hovered. The \c down.hovered property was introduced in QtQuick.Controls 2.1.
 
     \sa decrease()
 */
@@ -552,30 +661,106 @@ QQuickSpinButton *QQuickSpinBox::down() const
     return d->down;
 }
 
+/*!
+    \since QtQuick.Controls 2.2 (Qt 5.9)
+    \qmlproperty flags QtQuick.Controls::SpinBox::inputMethodHints
+
+    This property provides hints to the input method about the expected content
+    of the spin box and how it should operate.
+
+    The default value is \c Qt.ImhDigitsOnly.
+
+    \include inputmethodhints.qdocinc
+*/
+Qt::InputMethodHints QQuickSpinBox::inputMethodHints() const
+{
+    Q_D(const QQuickSpinBox);
+    return d->inputMethodHints;
+}
+
+void QQuickSpinBox::setInputMethodHints(Qt::InputMethodHints hints)
+{
+    Q_D(QQuickSpinBox);
+    if (d->inputMethodHints == hints)
+        return;
+
+    d->inputMethodHints = hints;
+    emit inputMethodHintsChanged();
+}
+
+/*!
+    \since QtQuick.Controls 2.2 (Qt 5.9)
+    \qmlproperty bool QtQuick.Controls::SpinBox::inputMethodComposing
+    \readonly
+
+    This property holds whether an editable spin box has partial text input from an input method.
+
+    While it is composing, an input method may rely on mouse or key events from the spin box to
+    edit or commit the partial text. This property can be used to determine when to disable event
+    handlers that may interfere with the correct operation of an input method.
+*/
+bool QQuickSpinBox::isInputMethodComposing() const
+{
+    Q_D(const QQuickSpinBox);
+    return d->contentItem && d->contentItem->property("inputMethodComposing").toBool();
+}
+
 /*!
     \qmlmethod void QtQuick.Controls::SpinBox::increase()
 
-    Increases the value by \l stepSize.
+    Increases the value by \l stepSize, or \c 1 if stepSize is not defined.
 
     \sa stepSize
 */
 void QQuickSpinBox::increase()
 {
     Q_D(QQuickSpinBox);
-    setValue(d->value + d->effectiveStepSize());
+    d->increase(false);
 }
 
 /*!
     \qmlmethod void QtQuick.Controls::SpinBox::decrease()
 
-    Decreases the value by \l stepSize.
+    Decreases the value by \l stepSize, or \c 1 if stepSize is not defined.
 
     \sa stepSize
 */
 void QQuickSpinBox::decrease()
 {
     Q_D(QQuickSpinBox);
-    setValue(d->value - d->effectiveStepSize());
+    d->decrease(false);
+}
+
+void QQuickSpinBox::focusInEvent(QFocusEvent *event)
+{
+    Q_D(QQuickSpinBox);
+    QQuickControl::focusInEvent(event);
+
+    // When an editable SpinBox gets focus, it must pass on the focus to its editor.
+    if (d->editable && d->contentItem && !d->contentItem->hasActiveFocus())
+        d->contentItem->forceActiveFocus(event->reason());
+}
+
+void QQuickSpinBox::hoverEnterEvent(QHoverEvent *event)
+{
+    Q_D(QQuickSpinBox);
+    QQuickControl::hoverEnterEvent(event);
+    d->updateHover(event->posF());
+}
+
+void QQuickSpinBox::hoverMoveEvent(QHoverEvent *event)
+{
+    Q_D(QQuickSpinBox);
+    QQuickControl::hoverMoveEvent(event);
+    d->updateHover(event->posF());
+}
+
+void QQuickSpinBox::hoverLeaveEvent(QHoverEvent *event)
+{
+    Q_D(QQuickSpinBox);
+    QQuickControl::hoverLeaveEvent(event);
+    d->down->setHovered(false);
+    d->up->setHovered(false);
 }
 
 void QQuickSpinBox::keyPressEvent(QKeyEvent *event)
@@ -586,7 +771,7 @@ void QQuickSpinBox::keyPressEvent(QKeyEvent *event)
     switch (event->key()) {
     case Qt::Key_Up:
         if (d->upEnabled()) {
-            increase();
+            d->increase(true);
             d->up->setPressed(true);
             event->accept();
         }
@@ -594,7 +779,7 @@ void QQuickSpinBox::keyPressEvent(QKeyEvent *event)
 
     case Qt::Key_Down:
         if (d->downEnabled()) {
-            decrease();
+            d->decrease(true);
             d->down->setPressed(true);
             event->accept();
         }
@@ -620,51 +805,6 @@ void QQuickSpinBox::keyReleaseEvent(QKeyEvent *event)
     setAccessibleProperty("pressed", false);
 }
 
-bool QQuickSpinBox::childMouseEventFilter(QQuickItem *child, QEvent *event)
-{
-    Q_D(QQuickSpinBox);
-    switch (event->type()) {
-    case QEvent::MouseButtonPress:
-        return d->handleMousePressEvent(child, static_cast<QMouseEvent *>(event));
-    case QEvent::MouseMove:
-        return d->handleMouseMoveEvent(child, static_cast<QMouseEvent *>(event));
-    case QEvent::MouseButtonRelease:
-        return d->handleMouseReleaseEvent(child, static_cast<QMouseEvent *>(event));
-    case QEvent::UngrabMouse:
-        return d->handleMouseUngrabEvent(child);
-    default:
-        return false;
-    }
-}
-
-void QQuickSpinBox::mousePressEvent(QMouseEvent *event)
-{
-    Q_D(QQuickSpinBox);
-    QQuickControl::mousePressEvent(event);
-    d->handleMousePressEvent(this, event);
-}
-
-void QQuickSpinBox::mouseMoveEvent(QMouseEvent *event)
-{
-    Q_D(QQuickSpinBox);
-    QQuickControl::mouseMoveEvent(event);
-    d->handleMouseMoveEvent(this, event);
-}
-
-void QQuickSpinBox::mouseReleaseEvent(QMouseEvent *event)
-{
-    Q_D(QQuickSpinBox);
-    QQuickControl::mouseReleaseEvent(event);
-    d->handleMouseReleaseEvent(this, event);
-}
-
-void QQuickSpinBox::mouseUngrabEvent()
-{
-    Q_D(QQuickSpinBox);
-    QQuickControl::mouseUngrabEvent();
-    d->handleMouseUngrabEvent(this);
-}
-
 void QQuickSpinBox::timerEvent(QTimerEvent *event)
 {
     Q_D(QQuickSpinBox);
@@ -673,24 +813,25 @@ void QQuickSpinBox::timerEvent(QTimerEvent *event)
         d->startPressRepeat();
     } else if (event->timerId() == d->repeatTimer) {
         if (d->up->isPressed())
-            increase();
+            d->increase(true);
         else if (d->down->isPressed())
-            decrease();
+            d->decrease(true);
     }
 }
 
+#if QT_CONFIG(wheelevent)
 void QQuickSpinBox::wheelEvent(QWheelEvent *event)
 {
     Q_D(QQuickSpinBox);
     QQuickControl::wheelEvent(event);
     if (d->wheelEnabled) {
-        const int oldValue = d->value;
         const QPointF angle = event->angleDelta();
         const qreal delta = (qFuzzyIsNull(angle.y()) ? angle.x() : angle.y()) / QWheelEvent::DefaultDeltasPerStep;
-        setValue(oldValue + qRound(d->effectiveStepSize() * delta));
-        event->setAccepted(d->value != oldValue);
+        if (!d->setValue(d->value + qRound(d->effectiveStepSize() * delta), true))
+           event->ignore();
     }
 }
+#endif
 
 void QQuickSpinBox::componentComplete()
 {
@@ -710,9 +851,20 @@ void QQuickSpinBox::itemChange(ItemChange change, const ItemChangeData &value)
 
 void QQuickSpinBox::contentItemChange(QQuickItem *newItem, QQuickItem *oldItem)
 {
-    Q_UNUSED(oldItem);
-    if (newItem)
+    Q_D(QQuickSpinBox);
+    if (QQuickTextInput *oldInput = qobject_cast<QQuickTextInput *>(oldItem))
+        disconnect(oldInput, &QQuickTextInput::inputMethodComposingChanged, this, &QQuickSpinBox::inputMethodComposingChanged);
+
+    if (newItem) {
         newItem->setActiveFocusOnTab(true);
+#if QT_CONFIG(cursor)
+        if (d->editable)
+            newItem->setCursor(Qt::IBeamCursor);
+#endif
+
+        if (QQuickTextInput *newInput = qobject_cast<QQuickTextInput *>(newItem))
+            connect(newInput, &QQuickTextInput::inputMethodComposingChanged, this, &QQuickSpinBox::inputMethodComposingChanged);
+    }
 }
 
 QFont QQuickSpinBox::defaultFont() const
@@ -720,23 +872,39 @@ QFont QQuickSpinBox::defaultFont() const
     return QQuickControlPrivate::themeFont(QPlatformTheme::EditorFont);
 }
 
-#ifndef QT_NO_ACCESSIBILITY
+#if QT_CONFIG(accessibility)
 QAccessible::Role QQuickSpinBox::accessibleRole() const
 {
     return QAccessible::SpinBox;
 }
+
+void QQuickSpinBox::accessibilityActiveChanged(bool active)
+{
+    Q_D(QQuickSpinBox);
+    QQuickControl::accessibilityActiveChanged(active);
+
+    if (active)
+        setAccessibleProperty("editable", d->editable);
+}
 #endif
 
 class QQuickSpinButtonPrivate : public QObjectPrivate
 {
 public:
-    QQuickSpinButtonPrivate() : pressed(false), indicator(nullptr) { }
+    QQuickSpinButtonPrivate()
+        : pressed(false),
+          hovered(false),
+          indicator(nullptr)
+    {
+    }
+
     bool pressed;
+    bool hovered;
     QQuickItem *indicator;
 };
 
-QQuickSpinButton::QQuickSpinButton(QQuickSpinBox *parent) :
-    QObject(*(new QQuickSpinButtonPrivate), parent)
+QQuickSpinButton::QQuickSpinButton(QQuickSpinBox *parent)
+    : QObject(*(new QQuickSpinButtonPrivate), parent)
 {
 }
 
@@ -756,6 +924,22 @@ void QQuickSpinButton::setPressed(bool pressed)
     emit pressedChanged();
 }
 
+bool QQuickSpinButton::isHovered() const
+{
+    Q_D(const QQuickSpinButton);
+    return d->hovered;
+}
+
+void QQuickSpinButton::setHovered(bool hovered)
+{
+    Q_D(QQuickSpinButton);
+    if (d->hovered == hovered)
+        return;
+
+    d->hovered = hovered;
+    emit hoveredChanged();
+}
+
 QQuickItem *QQuickSpinButton::indicator() const
 {
     Q_D(const QQuickSpinButton);
@@ -768,18 +952,12 @@ void QQuickSpinButton::setIndicator(QQuickItem *indicator)
     if (d->indicator == indicator)
         return;
 
-    QQuickControl *control = qobject_cast<QQuickControl*>(d->parent);
-    if (control)
-        QQuickControlPrivate::get(control)->deleteDelegate(d->indicator);
-    else
-        delete d->indicator;
-
+    QQuickControlPrivate::destroyDelegate(d->indicator, d->parent);
     d->indicator = indicator;
 
     if (indicator) {
         if (!indicator->parentItem())
             indicator->setParentItem(static_cast<QQuickItem *>(parent()));
-        indicator->setAcceptedMouseButtons(Qt::LeftButton);
     }
     emit indicatorChanged();
 }
