@@ -1,6 +1,6 @@
 /****************************************************************************
 **
-** Copyright (C) 2016 The Qt Company Ltd.
+** Copyright (C) 2017 The Qt Company Ltd.
 ** Contact: http://www.qt.io/licensing/
 **
 ** This file is part of the Qt Quick Templates 2 module of the Qt Toolkit.
@@ -35,6 +35,7 @@
 ****************************************************************************/
 
 #include "qquickswipedelegate_p.h"
+#include "qquickswipedelegate_p_p.h"
 #include "qquickcontrol_p_p.h"
 #include "qquickitemdelegate_p_p.h"
 #include "qquickvelocitycalculator_p_p.h"
@@ -43,6 +44,9 @@
 #include <QtGui/private/qguiapplication_p.h>
 #include <QtGui/qpa/qplatformtheme.h>
 #include <QtQml/qqmlinfo.h>
+#include <QtQuick/private/qquickanimation_p.h>
+#include <QtQuick/private/qquicktransition_p.h>
+#include <QtQuick/private/qquicktransitionmanager_p_p.h>
 
 QT_BEGIN_NAMESPACE
 
@@ -53,15 +57,21 @@ QT_BEGIN_NAMESPACE
     \inqmlmodule QtQuick.Controls
     \since 5.7
     \ingroup qtquickcontrols2-delegates
-    \brief A swipable item delegate.
+    \brief Swipable item delegate.
 
     SwipeDelegate presents a view item that can be swiped left or right to
     expose more options or information. It is used as a delegate in views such
     as \l ListView.
 
-    SwipeDelegate inherits its API from AbstractButton. For instance, you can set
-    \l {AbstractButton::text}{text} and react to
-    \l {AbstractButton::clicked}{clicks} using the AbstractButton API.
+    In the following example, SwipeDelegate is used in a \l ListView to allow
+    items to be removed from it by swiping to the left:
+
+    \snippet qtquickcontrols2-swipedelegate.qml 1
+
+    SwipeDelegate inherits its API from \l ItemDelegate, which is inherited
+    from AbstractButton. For instance, you can set \l {AbstractButton::text}{text},
+    and react to \l {AbstractButton::clicked}{clicks} using the AbstractButton
+    API.
 
     Information regarding the progress of a swipe, as well as the components
     that should be shown upon swiping, are both available through the
@@ -90,33 +100,56 @@ QT_BEGIN_NAMESPACE
 
     \image qtquickcontrols2-swipedelegate-behind.gif
 
-    \sa {Customizing SwipeDelegate}, {Delegate Controls}
+    \sa {Customizing SwipeDelegate}, {Delegate Controls}, {Qt Quick Controls 2 - Swipe to Remove}{Swipe to Remove Example}
 */
 
 namespace {
+    typedef QQuickSwipeDelegateAttached Attached;
+
+    Attached *attachedObject(QQuickItem *item) {
+        return qobject_cast<Attached*>(qmlAttachedPropertiesObject<QQuickSwipeDelegate>(item, false));
+    }
+
     enum PositionAnimation {
         DontAnimatePosition,
         AnimatePosition
     };
 }
 
+class QQuickSwipeTransitionManager : public QQuickTransitionManager
+{
+public:
+    QQuickSwipeTransitionManager(QQuickSwipe *swipe);
+
+    void transition(QQuickTransition *transition, qreal position);
+
+protected:
+    void finished() override;
+
+private:
+    QQuickSwipe *m_swipe;
+};
+
 class QQuickSwipePrivate : public QObjectPrivate
 {
     Q_DECLARE_PUBLIC(QQuickSwipe)
 
 public:
-    QQuickSwipePrivate(QQuickSwipeDelegate *control) :
-        control(control),
-        positionBeforePress(0),
-        position(0),
-        wasComplete(false),
-        complete(false),
-        left(nullptr),
-        behind(nullptr),
-        right(nullptr),
-        leftItem(nullptr),
-        behindItem(nullptr),
-        rightItem(nullptr)
+    QQuickSwipePrivate(QQuickSwipeDelegate *control)
+        : control(control),
+          positionBeforePress(0),
+          position(0),
+          wasComplete(false),
+          complete(false),
+          enabled(true),
+          left(nullptr),
+          behind(nullptr),
+          right(nullptr),
+          leftItem(nullptr),
+          behindItem(nullptr),
+          rightItem(nullptr),
+          transition(nullptr),
+          transitionManager(nullptr)
     {
     }
 
@@ -136,6 +169,12 @@ public:
     void warnAboutMixingDelegates();
     void warnAboutSettingDelegatesWhileVisible();
 
+    bool hasDelegates() const;
+
+    bool isTransitioning() const;
+    void beginTransition(qreal position);
+    void finishTransition();
+
     QQuickSwipeDelegate *control;
     // Same range as position, but is set before press events so that we can
     // keep track of which direction the user must swipe when using left and right delegates.
@@ -145,6 +184,7 @@ public:
     // before the last press event.
     bool wasComplete;
     bool complete;
+    bool enabled;
     QQuickVelocityCalculator velocityCalculator;
     QQmlComponent *left;
     QQmlComponent *behind;
@@ -152,8 +192,37 @@ public:
     QQuickItem *leftItem;
     QQuickItem *behindItem;
     QQuickItem *rightItem;
+    QQuickTransition *transition;
+    QScopedPointer<QQuickSwipeTransitionManager> transitionManager;
 };
 
+QQuickSwipeTransitionManager::QQuickSwipeTransitionManager(QQuickSwipe *swipe)
+    : m_swipe(swipe)
+{
+}
+
+void QQuickSwipeTransitionManager::transition(QQuickTransition *transition, qreal position)
+{
+    qmlExecuteDeferred(transition);
+
+    QQmlProperty defaultTarget(m_swipe, QLatin1String("position"));
+    QQmlListProperty<QQuickAbstractAnimation> animations = transition->animations();
+    const int count = animations.count(&animations);
+    for (int i = 0; i < count; ++i) {
+        QQuickAbstractAnimation *anim = animations.at(&animations, i);
+        anim->setDefaultTarget(defaultTarget);
+    }
+
+    QList<QQuickStateAction> actions;
+    actions << QQuickStateAction(m_swipe, QLatin1String("position"), position);
+    QQuickTransitionManager::transition(actions, transition, m_swipe);
+}
+
+void QQuickSwipeTransitionManager::finished()
+{
+    QQuickSwipePrivate::get(m_swipe)->finishTransition();
+}
+
 QQuickSwipePrivate *QQuickSwipePrivate::get(QQuickSwipe *swipe)
 {
     return swipe->d_func();
@@ -168,7 +237,7 @@ QQuickItem *QQuickSwipePrivate::createDelegateItem(QQmlComponent *component)
     // the creation context will be null and we have to create it ourselves.
     if (!creationContext)
         creationContext = qmlContext(control);
-    QQmlContext *context = new QQmlContext(creationContext);
+    QQmlContext *context = new QQmlContext(creationContext, control);
     context->setContextObject(control);
     QQuickItem *item = qobject_cast<QQuickItem*>(component->beginCreate(context));
     if (item) {
@@ -270,7 +339,7 @@ void QQuickSwipePrivate::createLeftItem()
         Q_Q(QQuickSwipe);
         q->setLeftItem(createDelegateItem(left));
         if (!leftItem)
-            qmlInfo(control) << "Failed to create left item:" << left->errors();
+            qmlWarning(control) << "Failed to create left item:" << left->errors();
     }
 }
 
@@ -280,7 +349,7 @@ void QQuickSwipePrivate::createBehindItem()
         Q_Q(QQuickSwipe);
         q->setBehindItem(createDelegateItem(behind));
         if (!behindItem)
-            qmlInfo(control) << "Failed to create behind item:" << behind->errors();
+            qmlWarning(control) << "Failed to create behind item:" << behind->errors();
     }
 }
 
@@ -290,7 +359,7 @@ void QQuickSwipePrivate::createRightItem()
         Q_Q(QQuickSwipe);
         q->setRightItem(createDelegateItem(right));
         if (!rightItem)
-            qmlInfo(control) << "Failed to create right item:" << right->errors();
+            qmlWarning(control) << "Failed to create right item:" << right->errors();
     }
 }
 
@@ -328,16 +397,51 @@ void QQuickSwipePrivate::createAndShowRightItem()
 
 void QQuickSwipePrivate::warnAboutMixingDelegates()
 {
-    qmlInfo(control) << "cannot set both behind and left/right properties";
+    qmlWarning(control) << "cannot set both behind and left/right properties";
 }
 
 void QQuickSwipePrivate::warnAboutSettingDelegatesWhileVisible()
 {
-    qmlInfo(control) << "left/right/behind properties may only be set when swipe.position is 0";
+    qmlWarning(control) << "left/right/behind properties may only be set when swipe.position is 0";
 }
 
-QQuickSwipe::QQuickSwipe(QQuickSwipeDelegate *control) :
-    QObject(*(new QQuickSwipePrivate(control)))
+bool QQuickSwipePrivate::hasDelegates() const
+{
+    return left || right || behind;
+}
+
+bool QQuickSwipePrivate::isTransitioning() const
+{
+    return transitionManager && transitionManager->isRunning();
+}
+
+void QQuickSwipePrivate::beginTransition(qreal newPosition)
+{
+    Q_Q(QQuickSwipe);
+    if (!transition) {
+        q->setPosition(newPosition);
+        finishTransition();
+        return;
+    }
+
+    if (!transitionManager)
+        transitionManager.reset(new QQuickSwipeTransitionManager(q));
+
+    transitionManager->transition(transition, newPosition);
+}
+
+void QQuickSwipePrivate::finishTransition()
+{
+    Q_Q(QQuickSwipe);
+    q->setComplete(qFuzzyCompare(qAbs(position), 1.0));
+    if (complete)
+        emit q->opened();
+    else
+        emit q->closed();
+}
+
+QQuickSwipe::QQuickSwipe(QQuickSwipeDelegate *control)
+    : QObject(*(new QQuickSwipePrivate(control)))
 {
 }
 
@@ -370,6 +474,8 @@ void QQuickSwipe::setLeft(QQmlComponent *left)
         d->leftItem = nullptr;
     }
 
+    d->control->setFiltersChildMouseEvents(d->hasDelegates());
+
     emit leftChanged();
 }
 
@@ -402,6 +508,8 @@ void QQuickSwipe::setBehind(QQmlComponent *behind)
         d->behindItem = nullptr;
     }
 
+    d->control->setFiltersChildMouseEvents(d->hasDelegates());
+
     emit behindChanged();
 }
 
@@ -434,6 +542,8 @@ void QQuickSwipe::setRight(QQmlComponent *right)
         d->rightItem = nullptr;
     }
 
+    d->control->setFiltersChildMouseEvents(d->hasDelegates());
+
     emit rightChanged();
 }
 
@@ -544,26 +654,75 @@ void QQuickSwipe::setComplete(bool complete)
 
     d->complete = complete;
     emit completeChanged();
+    if (d->complete)
+        emit completed();
 }
 
-class QQuickSwipeDelegatePrivate : public QQuickItemDelegatePrivate
+bool QQuickSwipe::isEnabled() const
 {
-    Q_DECLARE_PUBLIC(QQuickSwipeDelegate)
+    Q_D(const QQuickSwipe);
+    return d->enabled;
+}
 
-public:
-    QQuickSwipeDelegatePrivate(QQuickSwipeDelegate *control) :
-        swipe(control)
-    {
-    }
+void QQuickSwipe::setEnabled(bool enabled)
+{
+    Q_D(QQuickSwipe);
+    if (enabled == d->enabled)
+        return;
 
-    bool handleMousePressEvent(QQuickItem *item, QMouseEvent *event);
-    bool handleMouseMoveEvent(QQuickItem *item, QMouseEvent *event);
-    bool handleMouseReleaseEvent(QQuickItem *item, QMouseEvent *event);
+    d->enabled = enabled;
+    emit enabledChanged();
+}
 
-    void resizeContent() override;
+QQuickTransition *QQuickSwipe::transition() const
+{
+    Q_D(const QQuickSwipe);
+    return d->transition;
+}
 
-    QQuickSwipe swipe;
-};
+void QQuickSwipe::setTransition(QQuickTransition *transition)
+{
+    Q_D(QQuickSwipe);
+    if (transition == d->transition)
+        return;
+
+    d->transition = transition;
+    emit transitionChanged();
+}
+
+void QQuickSwipe::open(QQuickSwipeDelegate::Side side)
+{
+    Q_D(QQuickSwipe);
+    if (qFuzzyCompare(qAbs(d->position), 1.0))
+        return;
+
+    if ((side != QQuickSwipeDelegate::Left && side != QQuickSwipeDelegate::Right)
+            || (!d->left && !d->behind && side == QQuickSwipeDelegate::Left)
+            || (!d->right && !d->behind && side == QQuickSwipeDelegate::Right))
+        return;
+
+    d->beginTransition(side);
+    d->wasComplete = true;
+    d->velocityCalculator.reset();
+    d->positionBeforePress = d->position;
+}
+
+void QQuickSwipe::close()
+{
+    Q_D(QQuickSwipe);
+    if (qFuzzyIsNull(d->position))
+        return;
+
+    d->beginTransition(0.0);
+    d->wasComplete = false;
+    d->positionBeforePress = 0.0;
+    d->velocityCalculator.reset();
+}
+
+QQuickSwipeDelegatePrivate::QQuickSwipeDelegatePrivate(QQuickSwipeDelegate *control)
+    : swipe(control)
+{
+}
 
 bool QQuickSwipeDelegatePrivate::handleMousePressEvent(QQuickItem *item, QMouseEvent *event)
 {
@@ -574,12 +733,29 @@ bool QQuickSwipeDelegatePrivate::handleMousePressEvent(QQuickItem *item, QMouseE
     // events will go through the regular channels (mousePressEvent()) until then.
     if (qFuzzyIsNull(swipePrivate->position)) {
         q->mousePressEvent(event);
+        // The press point could be incorrect if the press happened over a child item,
+        // so we correct it after calling the base class' mousePressEvent(), rather
+        // than having to duplicate its code just so we can set the pressPoint.
+        pressPoint = item->mapToItem(q, event->pos());
         return true;
     }
 
+    // The position is non-zero, this press could be either for a delegate or the control itself
+    // (the control can be clicked to e.g. close the swipe). Either way, we must begin measuring
+    // mouse movement in case it turns into a swipe, in which case we grab the mouse.
     swipePrivate->positionBeforePress = swipePrivate->position;
     swipePrivate->velocityCalculator.startMeasuring(event->pos(), event->timestamp());
     pressPoint = item->mapToItem(q, event->pos());
+
+    // When a delegate uses the attached properties and signals, it declares that it wants mouse events.
+    Attached *attached = attachedObject(item);
+    if (attached) {
+        attached->setPressed(true);
+        // Stop the event from propagating, as QQuickItem explicitly ignores events.
+        event->accept();
+        return true;
+    }
+
     return false;
 }
 
@@ -592,12 +768,17 @@ bool QQuickSwipeDelegatePrivate::handleMouseMoveEvent(QQuickItem *item, QMouseEv
             stopPressAndHold();
     }
 
+    // The delegate can still be pressed when swipe.enabled is false,
+    // but the mouse moving shouldn't have any effect on swipe.position.
+    QQuickSwipePrivate *swipePrivate = QQuickSwipePrivate::get(&swipe);
+    if (!swipePrivate->enabled)
+        return false;
+
     // Protect against division by zero.
     if (width == 0)
         return false;
 
     // Don't bother reacting to events if we don't have any delegates.
-    QQuickSwipePrivate *swipePrivate = QQuickSwipePrivate::get(&swipe);
     if (!swipePrivate->left && !swipePrivate->right && !swipePrivate->behind)
         return false;
 
@@ -605,18 +786,22 @@ bool QQuickSwipeDelegatePrivate::handleMouseMoveEvent(QQuickItem *item, QMouseEv
     if (item == q && !pressed)
         return false;
 
-    const qreal distance = (event->pos() - pressPoint).x();
+    const QPointF mappedEventPos = item->mapToItem(q, event->pos());
+    const qreal distance = (mappedEventPos - pressPoint).x();
     if (!q->keepMouseGrab()) {
-        // Taken from QQuickDrawer::handleMouseMoveEvent; see comments there.
+        // Taken from QQuickDrawerPrivate::grabMouse; see comments there.
         int threshold = qMax(20, QGuiApplication::styleHints()->startDragDistance() + 5);
         const bool overThreshold = QQuickWindowPrivate::dragOverThreshold(distance, Qt::XAxis, event, threshold);
         if (window && overThreshold) {
             QQuickItem *grabber = q->window()->mouseGrabberItem();
             if (!grabber || !grabber->keepMouseGrab()) {
                 q->grabMouse();
-                q->setKeepMouseGrab(overThreshold);
+                q->setKeepMouseGrab(true);
                 q->setPressed(true);
                 swipe.setComplete(false);
+
+                if (Attached *attached = attachedObject(item))
+                    attached->setPressed(false);
             }
         }
     }
@@ -665,8 +850,17 @@ bool QQuickSwipeDelegatePrivate::handleMouseMoveEvent(QQuickItem *item, QMouseEv
                 position = distance > 0 ? normalizedDistance - 1.0 : normalizedDistance + 1.0;
             }
 
+            if (swipePrivate->isTransitioning())
+                swipePrivate->transitionManager->cancel();
             swipe.setPosition(position);
         }
+    } else {
+        // The swipe wasn't initiated.
+        if (event->pos().y() < 0 || event->pos().y() > height) {
+            // The mouse went outside the vertical bounds of the control, so
+            // we should no longer consider it pressed.
+            q->setPressed(false);
+        }
     }
 
     event->accept();
@@ -676,7 +870,7 @@ bool QQuickSwipeDelegatePrivate::handleMouseMoveEvent(QQuickItem *item, QMouseEv
 
 static const qreal exposeVelocityThreshold = 300.0;
 
-bool QQuickSwipeDelegatePrivate::handleMouseReleaseEvent(QQuickItem *, QMouseEvent *event)
+bool QQuickSwipeDelegatePrivate::handleMouseReleaseEvent(QQuickItem *item, QMouseEvent *event)
 {
     Q_Q(QQuickSwipeDelegate);
     QQuickSwipePrivate *swipePrivate = QQuickSwipePrivate::get(&swipe);
@@ -685,24 +879,44 @@ bool QQuickSwipeDelegatePrivate::handleMouseReleaseEvent(QQuickItem *, QMouseEve
     const bool hadGrabbedMouse = q->keepMouseGrab();
     q->setKeepMouseGrab(false);
 
+    // Animations for the background and contentItem delegates are typically
+    // only enabled when !control.down, so that the animations aren't running
+    // when the user is swiping. To ensure that the animations are enabled
+    // *before* the positions of these delegates change (via the swipe.setPosition() calls below),
+    // we must cancel the press. QQuickAbstractButton::mouseUngrabEvent() does this
+    // for us, but by then it's too late.
+    if (hadGrabbedMouse) {
+        // TODO: this is copied from QQuickAbstractButton::mouseUngrabEvent().
+        // Eventually it should be moved into a private helper so that we don't have to duplicate it.
+        q->setPressed(false);
+        stopPressRepeat();
+        stopPressAndHold();
+        emit q->canceled();
+    }
+
     // The control can be exposed by either swiping past the halfway mark, or swiping fast enough.
     const qreal swipeVelocity = swipePrivate->velocityCalculator.velocity().x();
     if (swipePrivate->position > 0.5 ||
         (swipePrivate->position > 0.0 && swipeVelocity > exposeVelocityThreshold)) {
-        swipe.setPosition(1.0);
-        swipe.setComplete(true);
+        swipePrivate->beginTransition(1.0);
         swipePrivate->wasComplete = true;
     } else if (swipePrivate->position < -0.5 ||
         (swipePrivate->position < 0.0 && swipeVelocity < -exposeVelocityThreshold)) {
-        swipe.setPosition(-1.0);
-        swipe.setComplete(true);
+        swipePrivate->beginTransition(-1.0);
         swipePrivate->wasComplete = true;
-    } else {
-        swipe.setPosition(0.0);
-        swipe.setComplete(false);
+    } else if (!swipePrivate->isTransitioning()) {
+        swipePrivate->beginTransition(0.0);
         swipePrivate->wasComplete = false;
     }
 
+    if (Attached *attached = attachedObject(item)) {
+        const bool wasPressed = attached->isPressed();
+        if (wasPressed) {
+            attached->setPressed(false);
+            emit attached->clicked();
+        }
+    }
+
     // Only consume child events if we had grabbed the mouse.
     return hadGrabbedMouse;
 }
@@ -715,7 +929,7 @@ static void warnIfHorizontallyAnchored(QQuickItem *item, const QString &itemName
     QQuickAnchors *anchors = QQuickItemPrivate::get(item)->_anchors;
     if (anchors && (anchors->fill() || anchors->centerIn() || anchors->left().item || anchors->right().item)
             && !item->property("_q_QQuickSwipeDelegate_warned").toBool()) {
-        qmlInfo(item) << QString::fromLatin1("SwipeDelegate: cannot use horizontal anchors with %1; unable to layout the item.").arg(itemName);
+        qmlWarning(item) << QString::fromLatin1("SwipeDelegate: cannot use horizontal anchors with %1; unable to layout the item.").arg(itemName);
         item->setProperty("_q_QQuickSwipeDelegate_warned", true);
     }
 }
@@ -739,40 +953,119 @@ void QQuickSwipeDelegatePrivate::resizeContent()
     }
 }
 
-QQuickSwipeDelegate::QQuickSwipeDelegate(QQuickItem *parent) :
-    QQuickItemDelegate(*(new QQuickSwipeDelegatePrivate(this)), parent)
+QQuickSwipeDelegate::QQuickSwipeDelegate(QQuickItem *parent)
+    : QQuickItemDelegate(*(new QQuickSwipeDelegatePrivate(this)), parent)
 {
-    setFiltersChildMouseEvents(true);
 }
 
+/*!
+    \since QtQuick.Controls 2.2 (Qt 5.9)
+    \qmlmethod void QtQuick.Controls::SwipeDelegate::swipe.open(enumeration side)
+
+    This method sets the \c position of the swipe so that it opens
+    from the specified \a side.
+
+    Available values:
+    \value SwipeDelegate.Left  The \c position is set to \c 1, which makes the swipe open
+                               from the left. Either \c swipe.left or \c swipe.behind must
+                               have been specified; otherwise the call is ignored.
+    \value SwipeDelegate.Right The \c position is set to \c -1, which makes the swipe open
+                               from the right. Either \c swipe.right or \c swipe.behind must
+                               have been specified; otherwise the call is ignored.
+
+    Any animations defined for the \l {Item::}{x} position of \l {Control::}{contentItem}
+    and \l {Control::}{background} will be triggered.
+
+    \sa swipe, swipe.close()
+*/
+
+/*!
+    \since QtQuick.Controls 2.1 (Qt 5.8)
+    \qmlmethod void QtQuick.Controls::SwipeDelegate::swipe.close()
+
+    This method sets the \c position of the swipe to \c 0. Any animations
+    defined for the \l {Item::}{x} position of \l {Control::}{contentItem}
+    and \l {Control::}{background} will be triggered.
+
+    \sa swipe, swipe.open()
+*/
+
+/*!
+    \since QtQuick.Controls 2.2 (Qt 5.9)
+    \qmlsignal void QtQuick.Controls::SwipeDelegate::swipe.opened()
+
+    This signal is emitted when the delegate has been swiped open
+    and the transition has finished.
+
+    It is useful for performing some action upon completion of a swipe.
+    For example, it can be used to remove the delegate from the list
+    that it is in.
+
+    \sa swipe, swipe.closed()
+*/
+
+/*!
+    \since QtQuick.Controls 2.2 (Qt 5.9)
+    \qmlsignal void QtQuick.Controls::SwipeDelegate::swipe.closed()
+
+    This signal is emitted when the delegate has been swiped to closed
+    and the transition has finished.
+
+    It is useful for performing some action upon cancellation of a swipe.
+    For example, it can be used to cancel the removal of the delegate from
+    the list that it is in.
+
+    \sa swipe, swipe.opened()
+*/
+
+/*!
+    \since QtQuick.Controls 2.1 (Qt 5.8)
+    \qmlsignal void QtQuick.Controls::SwipeDelegate::swipe.completed()
+
+    This signal is emitted when \c swipe.complete becomes \c true.
+
+    It is useful for performing some action upon completion of a swipe.
+    For example, it can be used to remove the delegate from the list
+    that it is in.
+
+    \sa swipe
+*/
+
 /*!
     \qmlpropertygroup QtQuick.Controls::SwipeDelegate::swipe
     \qmlproperty real QtQuick.Controls::SwipeDelegate::swipe.position
     \qmlproperty bool QtQuick.Controls::SwipeDelegate::swipe.complete
+    \qmlproperty bool QtQuick.Controls::SwipeDelegate::swipe.enabled
     \qmlproperty Component QtQuick.Controls::SwipeDelegate::swipe.left
     \qmlproperty Component QtQuick.Controls::SwipeDelegate::swipe.behind
     \qmlproperty Component QtQuick.Controls::SwipeDelegate::swipe.right
     \qmlproperty Item QtQuick.Controls::SwipeDelegate::swipe.leftItem
     \qmlproperty Item QtQuick.Controls::SwipeDelegate::swipe.behindItem
     \qmlproperty Item QtQuick.Controls::SwipeDelegate::swipe.rightItem
+    \qmlproperty Transition QtQuick.Controls::SwipeDelegate::swipe.transition
 
     \table
     \header
-        \li Property
+        \li Name
         \li Description
     \row
         \li position
-        \li This property holds the position of the swipe relative to either
+        \li This read-only property holds the position of the swipe relative to either
             side of the control. When this value reaches either
             \c -1.0 (left side) or \c 1.0 (right side) and the mouse button is
             released, \c complete will be \c true.
     \row
         \li complete
-        \li This property holds whether the control is fully exposed after
+        \li This read-only property holds whether the control is fully exposed after
             having been swiped to the left or right.
 
             When complete is \c true, any interactive items declared in \c left,
             \c right, or \c behind will receive mouse events.
+    \row
+        \li enabled
+        \li This property determines whether or not the control can be swiped.
+
+            This property was added in QtQuick.Controls 2.2.
     \row
         \li left
         \li This property holds the left delegate.
@@ -804,25 +1097,33 @@ QQuickSwipeDelegate::QQuickSwipeDelegate(QQuickItem *parent) :
             \include qquickswipedelegate-interaction.qdocinc
     \row
         \li leftItem
-        \li This property holds the item instantiated from the \c left component.
+        \li This read-only property holds the item instantiated from the \c left component.
 
             If \c left has not been set, or the position hasn't changed since
             creation of the SwipeDelegate, this property will be \c null.
     \row
         \li behindItem
-        \li This property holds the item instantiated from the \c behind component.
+        \li This read-only property holds the item instantiated from the \c behind component.
 
             If \c behind has not been set, or the position hasn't changed since
             creation of the SwipeDelegate, this property will be \c null.
     \row
         \li rightItem
-        \li This property holds the item instantiated from the \c right component.
+        \li This read-only property holds the item instantiated from the \c right component.
 
             If \c right has not been set, or the position hasn't changed since
             creation of the SwipeDelegate, this property will be \c null.
+    \row
+        \li transition
+        \li This property holds the transition that is applied when a swipe is released,
+            or \l swipe.open() or \l swipe.close() is called.
+
+            \snippet qtquickcontrols2-swipedelegate-transition.qml 1
+
+            This property was added in Qt Quick Controls 2.2.
     \endtable
 
-    \sa {Control::}{contentItem}, {Control::}{background}
+    \sa {Control::}{contentItem}, {Control::}{background}, swipe.open(), swipe.close()
 */
 QQuickSwipe *QQuickSwipeDelegate::swipe() const
 {
@@ -830,6 +1131,11 @@ QQuickSwipe *QQuickSwipeDelegate::swipe() const
     return const_cast<QQuickSwipe*>(&d->swipe);
 }
 
+QQuickSwipeDelegateAttached *QQuickSwipeDelegate::qmlAttachedProperties(QObject *object)
+{
+    return new QQuickSwipeDelegateAttached(object);
+}
+
 static bool isChildOrGrandchildOf(QQuickItem *child, QQuickItem *item)
 {
     return item && (child == item || item->isAncestorOf(child));
@@ -860,6 +1166,14 @@ bool QQuickSwipeDelegate::childMouseEventFilter(QQuickItem *child, QEvent *event
         QMouseEvent *mouseEvent = static_cast<QMouseEvent *>(event);
         QQuickItemDelegate::mouseReleaseEvent(mouseEvent);
         return d->handleMouseReleaseEvent(child, mouseEvent);
+    } case QEvent::UngrabMouse: {
+        // If the mouse was pressed over e.g. rightItem and then dragged down,
+        // the ListView would eventually grab the mouse, at which point we must
+        // clear the pressed flag so that it doesn't stay pressed after the release.
+        Attached *attached = attachedObject(child);
+        if (attached)
+            attached->setPressed(false);
+        return false;
     } default:
         return false;
     }
@@ -871,7 +1185,11 @@ void QQuickSwipeDelegate::mousePressEvent(QMouseEvent *event)
 {
     Q_D(QQuickSwipeDelegate);
     QQuickItemDelegate::mousePressEvent(event);
+
     QQuickSwipePrivate *swipePrivate = QQuickSwipePrivate::get(&d->swipe);
+    if (!swipePrivate->enabled)
+        return;
+
     swipePrivate->positionBeforePress = swipePrivate->position;
     swipePrivate->velocityCalculator.startMeasuring(event->pos(), event->timestamp());
 }
@@ -879,14 +1197,24 @@ void QQuickSwipeDelegate::mousePressEvent(QMouseEvent *event)
 void QQuickSwipeDelegate::mouseMoveEvent(QMouseEvent *event)
 {
     Q_D(QQuickSwipeDelegate);
-    d->handleMouseMoveEvent(this, event);
+    if (filtersChildMouseEvents())
+        d->handleMouseMoveEvent(this, event);
+    else
+        QQuickItemDelegate::mouseMoveEvent(event);
 }
 
 void QQuickSwipeDelegate::mouseReleaseEvent(QMouseEvent *event)
 {
     Q_D(QQuickSwipeDelegate);
-    QQuickItemDelegate::mouseReleaseEvent(event);
-    d->handleMouseReleaseEvent(this, event);
+    if (!filtersChildMouseEvents() || !d->handleMouseReleaseEvent(this, event))
+        QQuickItemDelegate::mouseReleaseEvent(event);
+}
+
+void QQuickSwipeDelegate::touchEvent(QTouchEvent *event)
+{
+    // Don't allow QQuickControl accept the touch event, because QQuickSwipeDelegate
+    // is still based on synthesized mouse events
+    event->ignore();
 }
 
 void QQuickSwipeDelegate::geometryChanged(const QRectF &newGeometry, const QRectF &oldGeometry)
@@ -905,11 +1233,159 @@ QFont QQuickSwipeDelegate::defaultFont() const
     return QQuickControlPrivate::themeFont(QPlatformTheme::ListViewFont);
 }
 
-#ifndef QT_NO_ACCESSIBILITY
+#if QT_CONFIG(accessibility)
 QAccessible::Role QQuickSwipeDelegate::accessibleRole() const
 {
     return QAccessible::ListItem;
 }
 #endif
 
+class QQuickSwipeDelegateAttachedPrivate : public QObjectPrivate
+{
+    Q_DECLARE_PUBLIC(QQuickSwipeDelegateAttached)
+
+public:
+    QQuickSwipeDelegateAttachedPrivate() : pressed(false) { }
+
+    // True when left/right/behind is non-interactive and is pressed.
+    bool pressed;
+};
+
+/*!
+    \since QtQuick.Controls 2.1 (Qt 5.8)
+    \qmlattachedsignal QtQuick.Controls::SwipeDelegate::clicked()
+
+    This signal can be attached to a non-interactive item declared in
+    \c swipe.left, \c swipe.right, or \c swipe.behind, in order to react to
+    clicks. Items can only be clicked when \c swipe.complete is \c true.
+
+    For interactive controls (such as \l Button) declared in these
+    items, use their respective \c clicked() signal instead.
+
+    To respond to clicks on the SwipeDelegate itself, use its
+    \l {AbstractButton::}{clicked()} signal.
+
+    \note See the documentation for \l pressed for information on
+    how to use the event-related properties correctly.
+
+    \sa pressed
+*/
+
+QQuickSwipeDelegateAttached::QQuickSwipeDelegateAttached(QObject *object)
+    : QObject(*(new QQuickSwipeDelegateAttachedPrivate), object)
+{
+    QQuickItem *item = qobject_cast<QQuickItem *>(object);
+    if (item) {
+        // This allows us to be notified when an otherwise non-interactive item
+        // is pressed and clicked. The alternative is much more more complex:
+        // iterating through children that contain the event pos and finding
+        // the first one with an attached object.
+        item->setAcceptedMouseButtons(Qt::AllButtons);
+    } else {
+        qWarning() << "Attached properties of SwipeDelegate must be accessed through an Item";
+    }
+}
+
+/*!
+    \since QtQuick.Controls 2.1 (Qt 5.8)
+    \qmlattachedproperty bool QtQuick.Controls::SwipeDelegate::pressed
+    \readonly
+
+    This property can be attached to a non-interactive item declared in
+    \c swipe.left, \c swipe.right, or \c swipe.behind, in order to detect if it
+    is pressed. Items can only be pressed when \c swipe.complete is \c true.
+
+    For example:
+
+    \code
+    swipe.right: Label {
+        anchors.right: parent.right
+        height: parent.height
+        text: "Action"
+        color: "white"
+        padding: 12
+        background: Rectangle {
+            color: SwipeDelegate.pressed ? Qt.darker("tomato", 1.1) : "tomato"
+        }
+    }
+    \endcode
+
+    It is possible to have multiple items which individually receive mouse and
+    touch events. For example, to have two actions in the \c swipe.right item,
+    use the following code:
+
+    \code
+    swipe.right: Row {
+        anchors.right: parent.right
+        height: parent.height
+
+        Label {
+            id: moveLabel
+            text: qsTr("Move")
+            color: "white"
+            verticalAlignment: Label.AlignVCenter
+            padding: 12
+            height: parent.height
+
+            SwipeDelegate.onClicked: console.log("Moving...")
+
+            background: Rectangle {
+                color: moveLabel.SwipeDelegate.pressed ? Qt.darker("#ffbf47", 1.1) : "#ffbf47"
+            }
+        }
+        Label {
+            id: deleteLabel
+            text: qsTr("Delete")
+            color: "white"
+            verticalAlignment: Label.AlignVCenter
+            padding: 12
+            height: parent.height
+
+            SwipeDelegate.onClicked: console.log("Deleting...")
+
+            background: Rectangle {
+                color: deleteLabel.SwipeDelegate.pressed ? Qt.darker("tomato", 1.1) : "tomato"
+            }
+        }
+    }
+    \endcode
+
+    Note how the \c color assignment in each \l {Control::}{background} item
+    qualifies the attached property with the \c id of the label. This
+    is important; using the attached properties on an item causes that item
+    to accept events. Suppose we had left out the \c id in the previous example:
+
+    \code
+    color: SwipeDelegate.pressed ? Qt.darker("tomato", 1.1) : "tomato"
+    \endcode
+
+    The \l Rectangle background item is a child of the label, so it naturally
+    receives events before it. In practice, this means that the background
+    color will change, but the \c onClicked handler in the label will never
+    get called.
+
+    For interactive controls (such as \l Button) declared in these
+    items, use their respective \c pressed property instead.
+
+    For presses on the SwipeDelegate itself, use its
+    \l {AbstractButton::}{pressed} property.
+
+    \sa clicked()
+*/
+bool QQuickSwipeDelegateAttached::isPressed() const
+{
+    Q_D(const QQuickSwipeDelegateAttached);
+    return d->pressed;
+}
+
+void QQuickSwipeDelegateAttached::setPressed(bool pressed)
+{
+    Q_D(QQuickSwipeDelegateAttached);
+    if (pressed == d->pressed)
+        return;
+
+    d->pressed = pressed;
+    emit pressedChanged();
+}
+
 QT_END_NAMESPACE
